#!/opt/LifeKeeper/bin/perl
#
# Copyright (c) 2000-2022 SIOS Technology Corp.
#
# Description:  Base class for all Data Replication Recovery Kit scripts.
#
# Author:  tpd
#
######################## LifeKeeper Linux Change Log #######################
# lklin00003897 PRC  17May06 cleanup, add 2.6 kernel bitmap and async support
# lklin00003951 PRC  25Jul06 make sure bitmap gets removed during delete
# lklin00003961 PRC  28Jul06 use deadline I/O scheduler, allow nbd-server args
# lklin00003963 PRC  31Jul06 execute CreateFlagFile on sys, rather than locally
# lklin00003966 PRC  02Aug06 allow min and max speed limits to be set separately
# lklin00003985 PRC  11Aug06 make sure target goes OSU after forced resync
# lklin00004109 PRC  05Feb07 multiple target support
# lklin00004213 PRC  19Mar07 make sure data_corrupt gets set on remove
# lklin00004223 PRC  28Mar07 fix SplitBrain message
# lklin00004262 PRC  19Apr07 remove of inactive/OSF mirror should just return 0
# lklin00004006 PRC  14Sep07 rewind
# lklin00004353 PRC  25Sep07 fix instructions for split brain resolution
# lklin00004474 PRC  27Sep07 compression
# lklin00004486 PRC  10Oct07 remove logs on delete/unextend
# lklin00004501 PRC  22Oct07 read-only target mount by default
# lklin00004533 PRC  01Nov07 set I/O scheduler for target mount
# lklin00004576 PRC  12Dec07 use 64-bit ioctls to handle large devices (> 4TB)
# lklin00004666 PRC  06Mar08 add Xen SR Create functionality
# lklin00004661 PRC  14Mar08 avoid next outside loop block error
# lklin00004692 PRC  25Apr08 make sure dummy nbd device gets deleted
# lklin00004782 KAS   June08 Added SDRS support.
# lklin00004939 KAS  24Sep08 SDRS: new mirror types: 4,5,6
# lklin00004961 KAS  20Oct08 App_delete returns 0 on a delete instead of
#                            the return value from StopNetRaid().
# lklin00004976 PRC  28Oct08 avoid recover when system is down
# lklin00004981 jef  30Oct08 redirect lcdremexec stderr to /dev/null
# bug1          PRC  02Feb09 bitmap merging
# bug24         PRC  24Feb09 avoid full resyncs (clear flags more aggressively)
# bug26         PRC  24Feb09 substitute "-e" script name in messages
# bug85         ysa  18Mar09 Added GetLocalPartition
# bug18         PRC  27May09 fix flag functions to handle "/" in tag
# bug345        PRC  15Jun09 fix RecoverMirrors/ILLSTATE bug
# bug299        PRC  16Jun09 wait for remove to succeed during resume
# bug274        PRC  16Jun09 fix bitmap merge with no filesystem parent
# bug657        PRC  22Oct09 fix bitmap merge failure -- do a full resync
# bug706        PRC  29Oct09 fix typo "aready"
# bug802        PRC  03Dec09 verify integrity of mirrorinfo file on update
# bug823        cmr  22Dec09 do not remove pid file when remexec fails
# bug862        PRC  07Jan10 simplify GrowArray
# bug1187       ysa  20Jun10 discard modprobe output
# bug1347       ysa  24Nov10 Avoid multiple simultaneous re-sync requests.
# bug1393       ysa  20Dec10 retry to tunnel connection.
# bug1488       cew  19Jan11 handle relative links
# bug1512       PRC  01Feb11 rename instead of ln -f (fix mirrorinfo clobber)
# bug986        PRC  22Feb11 don't unmount a dummy md if it's there, just use it
# bug1674       PRC  01Apr11 correct fix for 986
# bug1675/1677  PRC  01Apr11 fix bitmap clearing/wait for LK_INITDONE on resync
# bug1678       PRC  04Apr11 allow bitmap merge to be retried if it fails
# bug1728       PRC  18May11 fix split brain msg if bitmap merge is possible
# bug1757       PRC  20May11 set rotational parameter for nbds
# bug1790       ysa  15Jun11 force module loading (in generic module)
# bug1847       jef  12Jul11 Add fallback modprobe if modprobe with force fails
# bug1979       mat  04Aug11 change LogError to LogMessage when StopNetRAID fail
# bug2260       yuh  01Nov11 handle ConnectNBD when target_part is null
# bug2718       stt  26apr12 Added common logging tag/act context function.
# bug3182       jef  20Sep12 Remove forced modprobe to prevent kernel tainting
# bug3282       jef  17Jan13 Check mdadm --stop return on mirror resume
# bug3309       yuh  29Jan13 Set speed limit whenever a resync starts
# bug3308       yuh  31Jan13 logging around full and partial resyncs
# bug3422       jef  08Feb13 Add --force to mdadm --grow command
# bug3423       yuh  20Feb13 Stop mirror only when applicable
# bug2805       jef  26Feb13 On full resync dirty all bits if merge supported
# bug2732       yuh  12Aug13 fix for hung resync issue in RHEL 6
# bug3665       jef  22Jul13 Add --force on mdadm add/re-add for > 2TB extends
# bug7050       ysa  19Dec14 Avoid raid-check
# bug7085       ysa  24Feb15 file flush after writing
# bug7085       ryo  18Mar15 flush and sync before close
# bug7007       ysa  23Mar15 Avoid full resync for multi target cluster
# bug3292       ysa  17Apr15 Use C-locale time string
# bug7108       ysa  18May15 mdstat parser fix
# bug3675       ysa  02Jul15 Increase DEFAULT_BITMAP_CHUNK for 256
# bug7122       ryo  10Jul15 Change initial value for same behavior as default without config
# bug7150       ryo  24Dec15 Change the order of shared source/target processing in App_quickCheck and ResyncNBD
# bug7135       ryo  26Aug16 logging enhancement (flag and mirrorinfo log)
# bug7430       ysa  01Nov18 New nbd ioctl support
# bug7512       PRC  16Oct18 prevent restore when related mirror is data corrupt
# bug7008       ysa  06Nov18 Improve the message for a paused mirror
# bug7543       PRC  19Nov18 display partition size for LVM LVs
# bug7538       cew  05Dec18 Wait for source before doing resync
# bug7591       cew  17Jan19 get 'previous_source' first when adding target, full resync if can't determine last source
# bug7595       cew  18Jan19 if 'source' file missing during in-service on source, set data_corrupt and fail
# bug7598       cew  25Jan19 put back 'last_owner' flag file for startup handling, update log msg ID's to be unique
# bug7629       cew  12Feb19 During App_remove() only change target 'source' files if 'percent' is 100
# bug7657       jbb  21Mar19 Add check for odd number of sectors on source during restore if running kernel >= 4.12
# bug7672       cew  25Mar19 handle 1 node cluster and allow force online when missing source file.
# bug7683       PRC  30Mar19 speed up nbd request handling by increasing nr_requests block queue setting
# bug7691       ryo  22May19 support UEK5 nbd driver
# bug7725       ysa  01Aug19 cleanup nbd process on error
# bug7688       ryo  24Oct19 Update mdadm parsing
# PL-763        cew  02Dec19 Add default entry to not configure NBD during restore
# PL-3450       ryo  05Dec19 Skip operations during resizing
# PL-13245      jef  30Mar22 Throttle no ID message
package DRBase;

BEGIN {
	require '/etc/default/LifeKeeper.pl';
	use Exporter ();
	our ($VERSION, @ISA, @EXPORT, @EXPORT_OK, %EXPORT_TAGS);
	@ISA = qw(RKBase Exporter);
	@EXPORT = qw(&logmsg &logmsgdie &logmsgout &output &msg &msg2 &fatal
		&fatal2 &stack &stack_die &dbg &debug $compress $logdir $logmax
		$pingfail $pause $fullsync);
	%EXPORT_TAGS = ();
	@EXPORT_OK = qw($compress $logdir $logmax $pingfail $pause $fullsync);
}

use POSIX 'uname';
use Fcntl ':flock';
use Display;
use RKBase;
use lib "$ENV{'LKROOT'}/lkadm/subsys/gen/filesys/bin";
use filesysBase;
use DRLoadModule;
use Cwd;
use LKIP;
use IO::Handle;

my %wait_for_source_version = ("major" => 9, "minor" => 3, "point" => 2, "build" => 6830);

# LK::flg_create('', 'debug_dr'); # enable debug

use strict;

# CONSTANTS
#-------------------------------------------------------------------------
# Instance creation states - to be used as an enumerated type by creation
# and backout methods.
*InsState = \{
	NetRAIDCreated => 50,
	NetRAIDExtended => 75,
	Complete => 0,
};

my $XEN = "$ENV{'LKROOT'}/lkadm/subsys/steeleyeVM/XenVM/bin"; # Xen kit path

# Data Replication Recovery Kit Paths
our ($NETRAID_BIN, $NETRAID_ALARM);
*NETRAID_BIN = \( "$ENV{'LKROOT'}/lkadm/subsys/scsi/netraid/bin" );
*NETRAID_ALARM = \( "$ENV{'LKROOT'}/subsys/scsi/resources/netraid" );

foreach my $dir ('/usr/local/bin', $NETRAID_BIN) {
	$ENV{'PATH'} = "$dir:$ENV{'PATH'}" if ($ENV{'PATH'} !~ /\Q$dir\E/);
}

$ENV{'LKDR_TMPDIR'} ||= '/tmp'; # location for temp files (e.g., logs)

# LKDR network connection ping variables
$ENV{'LKDR_PING_NUMBER'} ||= 5;
$ENV{'LKDR_PING_TIMEOUT'} ||= 1;

$ENV{'LKDR_WAIT_FOR_PREVIOUS_SOURCE_TIMEOUT'} = -1 if (!defined($ENV{'LKDR_WAIT_FOR_PREVIOUS_SOURCE_TIMEOUT'}));
$ENV{'LKDR_CONNECT_NBD_DURING_RESTORE'} = "true" if (!defined($ENV{'LKDR_CONNECT_NBD_DURING_RESTORE'}));
$ENV{'LKDR_WAIT_TO_RESYNC'} = "filesys" if (!defined($ENV{'LKDR_WAIT_TO_RESYNC'}));

$ENV{'LKDR_MAX_STOP_WAIT'} ||= 24; # 5 second intervals

# LKDR resync speed limit
our ($PROC_SPEED_LIMIT_MIN_PATH, $PROC_SPEED_LIMIT_MAX_PATH);
*PROC_SPEED_LIMIT_MIN_PATH = \("/proc/sys/dev/raid/speed_limit_min");
*PROC_SPEED_LIMIT_MAX_PATH = \("/proc/sys/dev/raid/speed_limit_max");

our $DEFAULT_BITMAP_CHUNK;
*DEFAULT_BITMAP_CHUNK = \( 256 );

# SDRS: string appended to SDR resource ID to make it unique
our $NETRAID_ID_SUFFIX;
*NETRAID_ID_SUFFIX = \("-lksdr");

our $REC_LOCK_TEMPLATE;
*REC_LOCK_TEMPLATE = \("!volatile!%s-recovery");

# raid1 arrays can have MAX_TARGETS devices in them, as follows:
#      sysA   sysB   sysC   ...   sysH
# md0: nbd0   nbd1   nbd2   ...   nbd7
# md1: nbd8   nbd9   nbd10  ...   nbd15
#
# so, when A is the source:
#	md0: <localdisk_A> nbd1 nbd2 ...
# when B is the source:
#	md0: <localdisk_B> nbd0 nbd2 ...

# < 2.6.25: limit is 16 SDR mirrors per system with 8 disks each (128 = MAX_NBD)
our $MAX_TARGETS = $ENV{'LKDR_MAX_TARGETS'} || 8;

# setting names - used to get/set mirrorinfo settings (see Get/SetMirrorInfo)
our $compress = 'compress';
our $logdir = 'logdir';
our $logmax = 'logmax';
our $pause = 'pause';
our $pingfail = 'pingfail';
our $fullsync = 'fullsync';
our $group = 'group';
our $type = 'type';
our $logLevel_SetMirrorInfo = LK::LK_INFO;

$0 =~ s@.*/@@; # get basename

# log to stderr and the logfile by default
my $_logs = { 'STDERR' => 1, 'LOGFILE' => 1 };

DRLoadModule::LoadAllKernelModules(); # load all required kernel modules

my $_debug = (LK::flg_test('', 'debug_dr') == 0); # OPTIMIZATION

if ($_debug) {
	system("which mdadm >&2"); system("mdadm -V >&2");
}

# METHODS

sub debug {
	return $_debug;
}

sub logging { # params: \%hash
	my $hash = shift;
	$_logs = $hash;
}

sub mysprintf {
	return @_ if (@_ < 1);
	my $fmt = shift;
	return sprintf($fmt, @_);
}

sub calledby { # return $0 if it's not '-e', otherwise return calling subroutine
	return $0 if ($0 ne '-e');
	my $level = 2; # 2 means the caller of the caller
	# skip over message helper functions to the "real" calling routine
	$level++ while (((caller($level))[3] =~ /(dbg|output|msg|fatal)2*$/));
	return (caller($level))[3];
}

sub msg { # deprecated, use logmsg
	logmsg(LK::LK_INFO, '', '', 104085, @_);
}

sub logmsg {
	my $sev = shift;
	my $act = shift;
	my $tag = shift;
	my $id = shift;
	my $s = mysprintf(@_);
	LKBase::LogMsg($_logs, $sev, 'dr', $act, $tag, $id, $s);
}

sub logmsg2 {
	my $save = $_logs;
	logging( { 'LOGFILE' => 1 } );
	logmsg(@_);
	logging($save);
}

sub logmsgdie {
	logmsg(LK::LK_FATAL, @_);
	exit 1;
}

sub msg2 { # deprecated
	my $msg = shift;
	my $caller = calledby();
	LKBase::LogMessage($_logs, $caller, 'DR', $msg, @_);
}

sub output { # deprecated, use logmsgout
	my $s = mysprintf(@_);
	my $save = $_logs;
	print "$s\n";
	logging( { 'LOGFILE' => 1 } );
	logmsg(LK::LK_INFO, '', '', 104000, $s);
	logging($save);
}

sub logmsgout {
	my $sev = shift;
	my $act = shift;
	my $tag = shift;
	my $id = shift;
	my $s = mysprintf(@_);
	LKBase::LogMsg({ 'STDOUT' => 1, 'LOGFILE' => 1 }, $sev, 'dr', $act, $tag, $id, $s);
}

sub dbg {
	return if (!$_debug);
	logmsg(LK::LK_DEBUG, '', '', 104001, @_);
}

sub fatal { # deprecated, use logmsgdie
	msg(@_);
	exit 1;
}

sub fatal2 {
	msg2(@_);
	exit 1;
}

sub do_stack {
	my $die = shift;
	my $l = 0;
	my $msg = ($_[0] ? "@_" : ($die ? "Died" : "Error")) . " at ";

	while (++$l) {
		@_ = caller($l);
		last if (!defined($_[0]));
		$_[1] =~ s@^.*/@@;
		$msg .= ", " if ($l > 1);
		$msg .= "$_[1]:$_[2] ($_[3])";
	}

	if ($die) {
		logmsg(LK::LK_FATAL, '', '', 104002, $msg);
		die("\n");
	}

	dbg $msg;
}

sub stack {
	do_stack(0, @_);
}

sub stack_die {
	do_stack(1, @_);
}

# initialize (or re-initialize) members of self
sub init { # params: $self

	my $self = shift;

	if ($self->Val('Typ') ne 'netraid') { # sanity check
		logmsgdie('', $self->Val('Tag'), 104003, $self->Val('Tag') . " is not an SDR resource");
	}

	$self->{'id'} = $self->Val('ID');
	$self->{'tag'} = $self->Val('Tag');
	$self->{'sys'} = $self->Val('SysName');
	$self->{'state'} = $self->Val('State');

	my @info = split /\x02/, $self->Val('Info');
	stack "init: no info!" if (@info == 0);
	return if (@info == 0);
	shift @info if ($info[0] eq 'DELETEME'); # DELETEME is set during delete
	$self->{'num'} = $info[0];
	$self->{'md'} = "/dev/md$info[0]";
	$self->{'ip'} = $info[3]; # use only to determine if re-extend occurred
	$self->{'size'} = $info[4];
	$self->{'async'} = $info[7];
	$self->{'bitmap'} = $info[8];
	$self->{'chunk'} = $info[9];
	chomp($self->{'part'} = `getId -s $self->{'sys'} -p \"$self->{'id'}\"`);

	# upgrade? -- 'M' means resource has a mirrorinfo file
	$self->upgrade(@info) if ($info[10] !~ /M/);

	# per-target information
	$self->{'eqv'} = $self->GetMirrorInfo();
	# when this system is a target, which nbd device is used?
	$self->{'nbdnum'} = $self->{'eqv'}->{$self->{'sys'}}->{'num'};
	$self->{'target'} = $self->{'eqv'}->{$self->{'sys'}};
	$self->{'group'} = $self->{'eqv'}->{$self->{'sys'}}->{'group'};
	$self->{'type'} = $self->{'eqv'}->{$self->{'sys'}}->{'type'};
	$self->{'source'} = $self;

	my @eqv = LK::eqv_list($self->{'sys'}, $self->{'sys'}, $self->{'tag'});
	$self->{'numeqv'} = scalar(@eqv);
	foreach (@eqv) {
		split /\x01/;
		my $sys = $_[2];
		$self->{'eqv'}->{$sys}->{'tag'} = $_[3];
		$self->{'eqv'}->{$sys}->{'priority'} = $_[6];
		$self->{'eqv'}->{$sys}->{'state'} = LK::sys_getstate('', $sys);
		if ($self->{'group'}) {
			$self->{'groups'}->{$self->{'eqv'}->{$sys}->{'group'}} = -1;
		}
		# link each target back to the source object...
		$self->{'eqv'}->{$sys}->{'source'} = $self;
	}
	if ($self->{'group'}) {
		$self->{'groups'}->{$self->{'group'}} = -1;
		$self->{'numgroups'} = keys(%{$self->{'groups'}});
		$self->{'numdisks'} = $self->{'numgroups'};
		dbg "init: my group: " . $self->{'group'};
		dbg "init: groups: " . (join ', ', keys(%{$self->{'groups'}}));
	} else {
		$self->{'numdisks'} = $self->{'numeqv'} + 1;
	}
	dbg "init: numdisks " . $self->{'numdisks'};
	# tag and state don't get filled in above for self's target hash:
	$self->{'eqv'}->{$self->{'sys'}}->{'tag'} = $self->{'tag'};
}

# new(self)
#   Arguments:
#     self - reference to self which can be used to set additional member
#       variables.  self has to be passed as an argument since multiple
#       inheritance is being used.
#   Return Value:
#     Pointer to this class instance.
#
sub new {
	my $type = shift;
	my $self = shift;

	bless $$self, $type;

	init($$self);

	return $$self;
}

# modify the resources as necessary for upgrades:
#
# SDR 6.0 (pre-multiple target support, with no mirrorinfo file):
#     create mirrorinfo file, mark 11th info field with 'M'
sub upgrade { # params: self @info
	my $self = shift;
	my @info = shift;
	my @reminfo;
	my $remsys;
	my $remtag;
	my @eqv;

	logmsg(LK::LK_INFO, '', '', 104004, "Upgrading resource info for $self->{'tag'}...");
	return 0 if ($self->{'state'} !~ /^IS[PU]/); # upgrade from source only

	$self->Resource()->GetEqvIns(\@eqv);
	$remsys = $eqv[0]->SysName();
	$remtag = $eqv[0]->Tag();
	@reminfo = split /\x02/, (split /\x01/,
			LK::ins_list($remsys, '', '', '', $remtag, ''))[5];
	return 0 if (!@reminfo); # cannot complete upgrade - try again later

	logmsg(LK::LK_INFO, '', '', 104005, "Removing old target /dev/nbd$self->{'num'}...");
	system("mdadm --fail $self->{'md'} /dev/nbd$self->{'num'}");
	KillProc('nbd-client', $self->{'num'});
	system("mdadm --remove $self->{'md'} /dev/nbd$self->{'num'}");

	logmsg(LK::LK_INFO, '', '', 104006, "Creating mirrorinfo files...");
	DeleteMirrorInfoFile($self->{'num'});
	AddMirrorInfo('', $self->{'num'}, $self->{'sys'}, $self->{'num'} * $MAX_TARGETS, $reminfo[3], $self->{'async'}, 0, $self->{'state'} eq 'ISU' ? 1 : 0, '', 0, 0, 0, '', 0);
	AddMirrorInfo('', $self->{'num'}, $remsys, $self->{'num'} * $MAX_TARGETS + 1, $self->{'ip'}, $self->{'async'}, 0, 0, '', 0, 0, 0, '', 0);
	CopyMirrorInfo('', $self->{'num'}, $remsys);
	# mark the info field to indicate the upgrade has been completed
	$reminfo[10] = 'M';
	LK::ins_setinfo($remsys, $remtag, join "\x02", @reminfo);
	$info[10] = 'M';
	LK::ins_setinfo('', $self->{'tag'}, join "\x02", @info);
	logmsg(LK::LK_INFO, '', $self->{'tag'}, 104007, "Upgrade of resource info for $self->{'tag'} completed");
}

# sysfs helpers
sub mdsysfs { # params: mdnum [param]
	return "/sys/block/md$_[0]/md/$_[1]";
}

sub devsysfs { # params: mdnum device [param]
	return mdsysfs($_[0]) . "dev-$_[1]/$_[2]";
}

# helper for cat, stat, and grep
sub _filecmd { # params: cmd file [arg] [dest]
	my $cmd = shift;
	my $file = shift;
	my $arg = shift if ($cmd eq 'grep');
	my $dest = shift; # ($cmd eq 'grep') ? $_[3] : $_[2];
	my $data;

	dbg "$cmd $file $arg $dest";

	if ($dest) {
		$data=`lcdremexec -e -d "$dest" -- $cmd $arg $file 2>/dev/null`;
		if ($?) { return undef; }
		return $data;
	}

	return stat($file) if ($cmd eq 'stat');

	open(FD, "<$file") or return undef;
	while (<FD>) {
		if ($cmd eq 'cat') {
			$data .= $_;
		} elsif ($cmd eq 'grep') {
			$data .= $_ if ($_ =~ $arg);
		}
	}
	close(FD);
	return $data;
}

sub cat { # params: file [dest]
	return _filecmd('cat', $_[0], $_[1]);
}

sub rstat { # params: file [dest]
	return defined(_filecmd('stat', $_[0], $_[1])) ? 1 : undef;
}

sub rgrep { # params: file regex [dest]
	return _filecmd('grep', $_[0], $_[1], $_[2]);
}

# check for related (same hierarchy) data corrupt flags on the local system
# returns: undef if none found, otherwise the tag of the first data_corrupt found
sub RelatedDataCorruptExists { # params: self
	my $self = shift;

	my $hier = {};
	my $taglist = [$self->{'tag'}];
	LKBase::GetHierarchy('', $taglist, $hier);

	foreach my $tag (keys %$hier) {

		next if ($tag eq $self->{'tag'}); # skip if it's us, there's a separate check

		if (-f "$ENV{'LKROOT'}/subsys/scsi/resources/netraid/${tag}_data_corrupt") {
			return $tag; # found
		}
	}
	return undef; # not found
}

sub bitmap_set_bits { # params: mdnum ranges
	my $num = shift;
	my $ranges = shift;
	my $ret = 0;

	dbg "md$num: bitmap_set_bits($ranges)";
	open (FD, '>' . mdsysfs($num, 'bitmap_set_bits')) or return $!;
	print FD $ranges or $ret = $!;
	close(FD);

	return $ret;
}

sub dirty_bits { # params: self
	my $self = shift;

	my @out = `mdadm -X $self->{'bitmap'}`;
	while ($_ = pop @out) { # optimized: last line is Bitmap
		/Bitmap\s*:\s*.*,\s*(\d+)\s*dirty.*/;
		if (defined($1)) {
			return $1;
		}
	}
	return 0;
}

# return 1 if merge fails and full resync will be forced
# return -1 if merge fails and the merge will be retried later
sub MergeBitmap { # params: self target
	my $self = shift;
	my $target = shift;

	logmsg(LK::LK_INFO, '', '', 104008, "$self->{'md'}: merging bitmap from target \"$target->{'sys'}\"");

	my $bitmap = (split /\x02/, (split /\x01/,
	    LK::ins_list($target->{'sys'}, '', '', '', $target->{'tag'}, '')
	    )[5])[8];

	my $ranges = `lcdremexec -d $target->{'sys'} -- bitmap -m $bitmap`;
	if ($? == 0) {
		if (bitmap_set_bits($self->{'num'}, $ranges) == 0) {
			my $dirty = $self->dirty_bits();
			$dirty *= ($self->{'chunk'} * 1024);
			logmsg(LK::LK_INFO, '', '', 104009, "$self->{'md'}: bitmap merged, resyncing %s", HumanSize($dirty));
			return 0;
		}
	}

	# the merge has failed, determine what to do
	my $action = "will retry later. To force a full resync now: pause the mirror and force a full resync with the 'mirror_action' command.";
	my $retry = 1;
	if ($ENV{'LKDR_MERGE_NO_RETRY'}) {
		$action = "force full resync";
		$retry = 0;
	}

	if ($ranges) { # if bitmap command failed, output error
		logmsg(LK::LK_ERROR, '', '', 104106, "remote 'bitmap -m' command failed on $target->{'sys'}: $ranges");
	}
	logmsg(LK::LK_ERROR, '', '', 104010, "$self->{'md'}: bitmap merge failed, $action");

	return -1 if ($retry); # merge failed, retry later

	logmsg(LK::LK_INFO, '', $self->{'tag'}, 104179, "Set full resync to target %s.", $target->{'sys'});
	$target->{'fullsync'} = 1;
	SetMirrorInfo($target->{'sys'}, $self->{'tag'}, $fullsync, 1, 1);
	return 1; # merge failed, full resync required
}

sub SRUnconfig {
	my $self = shift;

	# Xen SR mirror -- unconfigure the storage repository
	system("$XEN/sr_unconfig -d $self->{'md'}");
}

sub SRUnavail { # params: md_dev [system]
	my $dev = shift;
	my $sys = shift;

	# Xen SR mirror -- mark the storage repository as unavailable
	system("lcdremexec -d \"$sys\" -e -- $XEN/sr_unavail -d $dev");
}

sub SRAvail { # params: md_dev [system]
	my $dev = shift;
	my $sys = shift;

	# Xen SR mirror -- mark the storage repository as available
	system("lcdremexec -d \"$sys\" -e -- $XEN/sr_avail -d $dev");
}

sub _MirrorAvail {
	my $num;
	my $type;
	my $sys;

	if (ref($_[0]) && $_[0] =~ /^\w+=HASH\(/) {
		my $self = shift;
		$num = $self->{'source'}->{'num'};
		$type = $self->{'type'};
		$sys = $self->{'sys'};
	} else {
		$num = shift;
		$type = shift;
		$sys = shift;
	}

	my $avail = shift;

	if (IsTypeXEN($type)) {
		SRAvail("/dev/md$num", $sys) if ($avail);
		SRUnavail("/dev/md$num", $sys) if (not $avail);
	}

	if (UpdateRaidCheck("md$num", $avail)) {
		my $guide = $avail?"add":"remove";
		logmsg(LK::LK_WARN, '', '', 104157, "/etc/sysconfig/raid-check update failed. Please %s \"md%d\" to SKIP_DEVS.", $guide, $num);
	}

}

sub MirrorAvail {
	_MirrorAvail(@_, 1);
}

sub MirrorUnavail {
	_MirrorAvail(@_, 0);
}

# LockRecovery
# Return non-zero if an instance of recovery is already running for our resource.
# This is determined through the existence of a flag based on the currrent tag name.
# If no recovery is running, then create the 'lock' flag and return 0.
#
sub LockRecovery {
        my $tag = shift;
        my $ret = 0;

	# The lock flag is of the format  PREFIX.pid
        my $recFlagPrefix = sprintf "$REC_LOCK_TEMPLATE", $tag;
        my $thisRecFlag = $recFlagPrefix . ".$$";

        @_ = `$ENV{'LKROOT'}/bin/flg_list 2>/dev/null`;
        my @recInstances = grep /$recFlagPrefix/, @_;
        foreach my $inst (@recInstances) {
                chomp($inst);
                # for each flag that fits our flag template, look
                # to see if the pid actually exists; if so, then
                # a recovery is in progress.
                # If flags are found matching our template for which
                # the process no longer exists, then the flag
                # is removed for cleanup purposes.
                $_ = $inst =~ /$recFlagPrefix.(\d+)/;
                my $pid = $1;
                `ps $pid >/dev/null 2>&1`;
                if ($? == 0) {
                        # recovery in progress.
                        $ret = 1;
                } else {
                        LK::flg_remove('', $inst);
                }
        }

	# no recovery running - create the locking flag.
	if ($ret == 0) {
		LK::flg_create('', $thisRecFlag);
	}
        return $ret;
}

# UnlockRecovery
# Remove the lock flag created by LockRecovery() above.
sub UnlockRecovery {
	my $tag = shift;
        my $recFlagPrefix = sprintf "$REC_LOCK_TEMPLATE", $tag;

	LK::flg_remove('', "$recFlagPrefix.$$");
	return 0;
}

sub PreRemove {
	my $self = shift;
	$self->SRUnconfig() if ($self->IsXEN());
}

sub PreRestore {
	my $self = shift;
	$self->SRUnconfig() if ($self->IsXEN());
}

# remove last_owner flag on target system
# Xen: cleanup target system (remove dummy md, unplug PBD, etc.)
sub CleanupTarget {
	my $target = shift;

	$target->DeleteFlagFile('last_owner');

	return if (! $target->IsXEN());

	system("lcdremexec -e -d \"$target->{'sys'}\" -- mdadm --stop $target->{'md'} 2>/dev/null");
	$target->MirrorUnavail();
}

# return 0 on failure, 1 otherwise
sub DetermineResyncType { # params: self
	my $self = shift;
	my @modestr = ("Partial", "Full");

	my $config = $self->GetClusterConfig();
	dbg "cluster: $config";
	if ($config == 2) {
		# only worry about last source when multi-target
		my $previous_source = $self->ReadFlagFile('source');
		if ($previous_source eq '' && $ENV{'SDR_FORCE_ONLINE'}) {
			logmsg(LK::LK_INFO, '', $self->{'tag'}, 104225, "The \"source\" flag file on %s to track mirror source does not exist.  Force online with a full resync to all targets.", $self->{'sys'});
			foreach ($self->SystemKeys()) {
				my $target = $self->Target($_);
				next if (!defined($target) or $target->{'sys'} eq '');
				$target->{'fullsync'} = 1;
				$self->SetFullSync($target->{'sys'});
				dbg "$self->{'md'}: full resync $target->{'sys'}";
			}
		} elsif ($previous_source eq '') {
			logmsg(LK::LK_ERROR, '', $self->{'tag'}, 104176, "The \"source\" flag file on %s to track mirror source does not exist.  Full resync is required.", $self->{'sys'});
			if (!$self->CreateFlagFile('data_corrupt')) {
				logmsg(LK::LK_ERROR, '', $self->{'tag'}, 104207, 'Failed to create "data_corrupt" flag file on "%s".', $self->{'sys'});
			}
			return 0;
		} elsif ($previous_source ne $self->{'sys'}) {
			if (!$self->CreateFlagFile('previous_source', $previous_source)) {
				logmsg(LK::LK_ERROR, '', $self->{'tag'}, 104224, "Failed to create \"previous_source\" flag file.");
				return 0;
			}
			# As targets are connected if they need a bitmap from a previous
			# source to be merged first then the 'previous_source' flag file
			# will be created.  This is created here mainly for the timestamp
			# for the timeout.  If/when this fails when a target is connected
			# then an appropriate error is logged.
			logmsg(LK::LK_INFO, '', $self->{'tag'}, 104183, "Mirror was last in-service on %s, merge its bitmap before resyncing targets.", $previous_source);
		}
	}
	if (!$self->CreateFlagFile('source', $self->{'sys'})) {
		logmsg(LK::LK_ERROR, '', $self->{'tag'}, 104214, "Failed to create \"source\" flag file on %s to track mirror source.", $self->{'sys'});
		return 0;
	}

	return 1;
}

# Get Cluster configuration (1x1 / multi site / multi target)
# return 0 (1x1), 1 (multi site), 2 (multi target)
sub GetClusterConfig { # params: self
	my $self = shift;
	my @syslist = $self->SystemKeys();
	push @syslist, $self->{'sys'};

	my %group;
	my $nonshared = 0;
	foreach my $sys (@syslist) {
		$_ = $self->{'eqv'}->{$sys}->{'group'};
		($_ ne '')?$group{$_}++:$nonshared++;
	}

	foreach my $sys (keys(%group)) {
		if ($group{$sys} == 1) {
			delete($group{$sys});
			$nonshared++;
		}
	}

	my $shared = keys(%group);
	dbg "shared $shared, nonshared $nonshared";
	if ($shared == 0 && $nonshared <= 2) {
		return 0;	# 1x1
	} elsif ($shared > 0 && $nonshared < 2) {
		return 1;	# multi site
	} else {
		return 2;	# multi target
	}
}

# routine: IsSystemReady
# description:
#   Is the system selected ALIVE and init done.
# return:
#  0 - not ready, ready otherwise
sub IsSystemReady { # params: self sys
	my $self = shift;
	my $sys = shift;

	my $target = $self->Target($sys);
	my $state = $target->{state};
	return 0 if ($state != $LK::ALIVE);
	return LKInitDone($sys);
}

sub FindNextTarget { # params: self pervious_source
	my $self = shift;
	my $exclude = shift;

	foreach my $sys ($self->SystemKeys()) {
		my $targ = $self->Target($sys);
		next if (!defined($targ) or $targ->{'sys'} eq '');
		next if ($targ->{'sys'} eq $exclude);
		next if (IsSharedSource($self, $targ));
		next if (CheckProc('nbd-client', $targ->{'num'})); # already configured
		return $sys;
	}
	dbg("FindNextTarget: did not find a target that was not $self->{'sys'} or $exclude or configured.");
	return '';
}

# routine: ContinueToWaitForPreviousSource
# description:
#   Continue to wait for the previous source if the previous source is not ready
#   and the timeout has not expired.
# return:
#  (0, '') - done waiting, nothing to merge
#  (0, <previous source>) - done waiting, previous source up before timeout, need to merge
#  (1, <previous source>) - continue waiting
sub ContinueToWaitForPreviousSource { # params: self
	my $self = shift;

	my $previous_source = $self->ReadFlagFile('previous_source');
	return (0, '') if ($previous_source eq '');
	if ($previous_source eq $self->{'sys'}) {
		# should not be possible
		logmsg(LK::LK_ERROR, '', $self->{'tag'}, 104180, "Internal Error: \"previous_source\" has the local system name (%s).", $previous_source);
		$self->DeleteFlagFile('previous_source');
		return (0, '');
	}
	return (0, $previous_source) if ($self->IsSystemReady($previous_source));

	my $target = $self->FindNextTarget($previous_source);
	if ($target eq '') {
		# should not be possible, all targets are connected
		logmsg(LK::LK_ERROR, '', $self->{'tag'}, 104181, "Internal Error: There are no targets waiting on %s to be merged.", $previous_source);
		$self->DeleteFlagFile('previous_source');
		return (0, '');
	}

	# LKDR_WAIT_FOR_PREVIOUS_SOURCE_TIMEOUT settings:
	#  -1 - wait indefinitely
	#   0 - do not wait any, start full resync immediately
	#  >0 - number of seconds to wait
	my $timeout = $ENV{'LKDR_WAIT_FOR_PREVIOUS_SOURCE_TIMEOUT'};
	if ($timeout == -1) {
		logmsg(LK::LK_EMERG, '', $self->{'tag'}, 104200, "Continue to wait for %s to merge bitmap and do partial resyncs to all targets, no timeout set.", $previous_source);
		logmsg(LK::LK_EMERG, '', $self->{'tag'}, 104201, "To stop waiting for the previous source (forcing a full resync to remaining waiting targets) run: \"%s/bin/mirror_action %s fullresync %s %s\" on %s.",
			$ENV{'LKROOT'}, $self->{'tag'}, $self->{'sys'}, $target, $self->{'sys'});
		return (1, $previous_source);
	}

	my $modtime = $self->FlagFileModificationTime('previous_source');
	if ($modtime == 0) {
		# was previous_source deleted between reading it above?
		logmsg(LK::LK_ERROR, '', $self->{'tag'}, 104177, "Failed to determine amount of time waiting on %s.", $previous_source);
		# this will end up aborting wait and do full resyncs
	}
	my $current = time();

	my $remaining = $modtime + $timeout - $current;
	if ($remaining > 0) {
		logmsg(LK::LK_EMERG, '', $self->{'tag'}, 104202, "Continue to wait for %s to merge bitmap and do partial resyncs to all targets. Continue to wait %d more seconds.", $previous_source, $remaining);
		logmsg(LK::LK_EMERG, '', $self->{'tag'}, 104203, "To stop waiting for the previous source (forcing a full resync to remaining waiting targets) run: \"%s/bin/mirror_action %s fullresync %s %s\" on %s.", $ENV{'LKROOT'}, $self->{'tag'}, $self->{'sys'}, $target, $self->{'sys'});
		return (1, $previous_source);
	}
	# TBD:  if timeout we could just do a full resync to one target and if the source returns
	# during the full resync then merge to do partial resyncs for rest.  > 3 node enhancement
	$self->CleanupWaitForPreviousSource($previous_source);
	logmsg(LK::LK_INFO, '', $self->{'tag'}, 104204, "Do not wait for %s to merge bitmap, begin full resync to other targets.", $previous_source) if ($timeout == 0);
	logmsg(LK::LK_INFO, '', $self->{'tag'}, 104220, "Exceeded timeout waiting for %s to merge bitmap, begin full resync to other targets.", $previous_source) if ($timeout != 0);
	return (0, '');
}

# routine: WaitToConnectNBD
# description:
#   Check if parent resources are in-service before allowing NBD's to be connected
# return:
#   0 - reconnect NBD
#   1 - wait to reconnect NBD
sub WaitToConnectNBD { # params: self
	my $self = shift;
	my $type = lc($ENV{'LKDR_WAIT_TO_RESYNC'});
	my $parent = '';
	my $state = '';

	return 0 if ($type eq 'false');

	# if any NBD's are already connected then do not wait to connect any others
	$self->UpdateStatus();
	foreach my $sys ($self->SystemKeys()) {
		my $target = $self->Target($sys);
		return 0 if ($target->{'percent'} > -1);
	}

	# check if appropriate parent resources are in-service
	$type = '' if ($type eq 'hierarchy');

	ParentNotInService($self->{'tag'}, $type, \$parent, \$state);
	return 0 if ($state !~ /^OS[UF]/);

	my $msg = '';
	$msg = " of type \"$type\"" if ($type ne '');

	# flag is used by res_state to show in GUI why the mirror is not connected
	LK::flg_create('', "!volatile!wait_to_resync_$self->{'tag'}");
	if ($state eq "OSF") {
		logmsg(LK::LK_EMERG, '', $self->{'tag'}, 104236, 'Resource "%s" is "OSF".  The mirror "%s" will wait to replicate data until all resources in the hierarchy%s are in-service.  This may indicate inconsistent data.  Verify the data is correct before replicating data; replication will continue when all resources in the hierarchy%s are in-service.  A full resync may be necessary (see "LKDR_WAIT_TO_RESYNC" in /etc/default/LifeKeeper).', $parent, $self->{'tag'}, $msg, $msg);
	} else {
		logmsg(LK::LK_WARN, '', $self->{'tag'}, 104237, 'Resource "%s" is "OSU".  The mirror "%s" will wait to replicate data until all resources in the hierarchy%s are in-service.  To replicate data immediately run: "%s/bin/mirror_action %s resume" on "%s" (see "LKDR_WAIT_TO_RESYNC" in /etc/default/LifeKeeper).', $parent, $self->{'tag'}, $msg, $ENV{'LKROOT'}, $self->{'tag'}, $self->{'sys'});
	}
	return 1;
}

# routine: CleanupWaitForPreviousSource
# description:
#   If the previous_source flag file exists delete it and mark all but the last
#   source and any already configured targets for full sync
# return:  none
sub CleanupWaitForPreviousSource { # params: self previous_source
	my $self = shift;
	my $previous_source = shift;

	return if (!$self->FlagFileExists('previous_source'));
	$previous_source = $self->ReadFlagFile('previous_source') if ($previous_source eq '');

	$self->MarkTargetsToFullsync($previous_source);
}

sub MarkTargetsToFullsync { # params: self exclude
	my $self = shift;
	my $exclude = shift;

	foreach my $targ ($self->SystemKeys()) {
		my $target = $self->Target($targ);
		next if (!defined($target) or $target->{'sys'} eq '');
		next if (defined($exclude) and $target->{'sys'} eq $exclude);
		next if (CheckProc('nbd-client', $target->{'num'})); # already configured
		logmsg(LK::LK_INFO, '', $self->{'tag'}, 104213, "Set full resync to target %s.", $target->{'sys'});
		$target->{'fullsync'} = 1;
		$self->SetFullSync($target->{'sys'});
	}
	$self->DeleteFlagFile('previous_source');
}

sub GetSyncStatus { # params: self
	my $self = shift;
	return $self->FlagFileExists('data_corrupt')?1:0;
}

sub now {
	my $date = `LC_TIME=C date +"%a %b %e %T %Z %Y"`;
	return $date;
}

sub is_in_resize {
	my $self = shift;
	my $action = shift;

	return 0 if (LK::flg_test('', "!in_resize!$self->{'tag'}") != 0);

	my $msg_level = LK::LK_ERROR;
	$msg_level = LK::LK_INFO if ($action eq 'quickCheck');
	logmsg($msg_level, $action, $self->{'tag'}, 104229, "An online resize operation is in progress. The $action action cannot be performed at this time.");
	return 1;
}

# App_restore
#   Description:  Called by the RKActionHandler::Restore method to do the
#     recovery kit specific restore work.
#   Arguments:  none.
#   Return Value:
#     0 if no errors, non-0 otherwise.
#
sub App_restore {
	my $self = shift;
	my $wait_secs = 0;
	my $max_wait_secs = 900;

	# in resize
	return 1 if ($self->is_in_resize('restore'));

	# Set nbd block layer params
	$self->SetBlockParams();

	# Wait for the lkstart check to finish
	if (! -f "$ENV{'LKROOT'}/config/LK_INITDONE") {
		while (! $self->FlagFileExists('startup_done') and
		    $wait_secs < $max_wait_secs) {
			$wait_secs++;
			sleep 1;
		}
		goto FAIL if (! $self->FlagFileExists('startup_done'));
	}
	$self->DeleteFlagFile('startup_done');

	# bug7657: If the kernel version is >= 4.12 and the partition has an
	# odd number of sectors, log an error and fail
	system("$NETRAID_BIN/is_odd_sector_size_supported $self->{'part'} > /dev/null 2>&1");
	if ($? != 0) {
		logmsg(LK::LK_ERROR, 'restore', $self->{'tag'}, 104132, 'The partition "%s" has an odd number of sectors and system "%s" is running kernel >= 4.12. Mirrors with this configuration will not work correctly with DataKeeper. Please see the SIOS product documentation for information on how to resize the mirror.', $self->{'part'}, $self->{'sys'});
		goto FAIL;
	}

	# if <tag>_data_corrupt exists either abort or override and log a warning
	if ($self->FlagFileExists('data_corrupt')) {
		if (!$ENV{'SDR_FORCE_ONLINE'}) {
			logmsg(LK::LK_ERROR, 'restore', $self->{'tag'}, 104086, 'The "%s_data_corrupt" flag is set in "%s/subsys/scsi/resources/netraid/" on system "%s". To avoid data corruption, LifeKeeper will not restore the resource.', $self->{'tag'}, $ENV{'LKROOT'}, $self->{'sys'});
			# don't dirty bitmap, it's overkill: $self->SetBitmap('dirty');
			goto FAIL;
		} else {
			# log a warning if the user overrides the data_corrupt flag for the mirror
			logmsg(LK::LK_WARN, 'restore', $self->{'tag'}, 104163, 'The "%s_data_corrupt" flag is set in "%s/subsys/scsi/resources/netraid/" on system "%s". The mirror is being forced online.', $self->{'tag'}, $ENV{'LKROOT'}, $self->{'sys'});
		}
	}

	# check for related mirror (in same hierarchy) data_corrupt flags
	if (my $rel_tag = $self->RelatedDataCorruptExists()) {
		if (!$ENV{'SDR_FORCE_RELATED_CORRUPT_ONLINE'}) {
			logmsg(LK::LK_ERROR, 'restore', $self->{'tag'}, 104164, 'The "%s_data_corrupt" flag for related mirror resource "%s" is set in "%s/subsys/scsi/resources/netraid/" on system "%s". To avoid data corruption, LifeKeeper will not restore this mirror or any related mirrors in the hierarchy.', $rel_tag, $rel_tag, $ENV{'LKROOT'}, $self->{'sys'});
			goto FAIL;
		} else {
			logmsg(LK::LK_ERROR, 'restore', $self->{'tag'}, 104165, 'The "%s_data_corrupt" flag for related mirror resource "%s" is set in "%s/subsys/scsi/resources/netraid/" on system "%s". The mirror resource "%s" is being forced online.',  $rel_tag, $rel_tag, $ENV{'LKROOT'}, $self->{'sys'}, $self->{'tag'});
		}
	}

	$self->PreRestore(); # do optional pre-restore steps

	$self->DeleteFlagFile('data_corrupt');

	# we've got a dummy md (see TempMount) so just use it
	$self->SetPause(0, $self->{'sys'}); # in case we were paused

	$self->UpdateStatus();

	# Discard old rewind log files
	CleanLogs($self->{'target'}->{'rewindlog'},
		  $self->{'target'}->{'tagfile'});

	if ($self->{'percent'} == -1) { # percent == -1 means no mirror exists

		# Create the mirror in degraded mode (so device size is correct)
		if (BuildArray($self->{'size'}, $self->{'part'}, $self->{'num'},
		    $self->{'numdisks'}, $self->{'async'}, $self->{'bitmap'},
		    $self->{'chunk'}) != 0)
		{
			goto FAIL;
		}
	}

	# determine if partial or full resyncs are needed to each target
	goto FAIL if (!$self->DetermineResyncType());

	# Create the last owner flag file. If another server is down,
	# it will check for this one when it comes back up.
	if (!$self->CreateFlagFile('last_owner')) {
		logmsg(LK::LK_ERROR, '', $self->{'tag'}, 104222, 'Failed to create "last_owner" flag file to track mirror source".  This may allow in-service of mirror on old data.');
	}

	my $previous_source = $self->ReadFlagFile('previous_source');
	foreach my $sys ($self->SystemKeys()) {
		my $target = $self->Target($sys);
		$target->DeleteFlagFile('last_owner');
		$self->SetPingFailCount(0, $sys);

		# set shared source systems to same state as local system
		if (IsSharedSource($self, $target)) {
			$target->DeleteFlagFile('data_corrupt');
			$target->{'fullsync'} = 0;
			SetMirrorInfo($target->{'sys'}, $self->{'tag'},
				$fullsync, 0, 1);
			next;
		}
		# set data corrupt on waiting targets
		next if ($target->{'sys'} eq $previous_source);
		next if ($target->{'state'} != $LK::ALIVE);
		if (!$target->CreateFlagFile('data_corrupt')) {
			logmsg(LK::LK_ERROR, '', $self->{'tag'}, 104211, 'Failed to create "data_corrupt" flag file on target "%s".', $target->{'sys'});
		}
	}

	if (lc($ENV{'LKDR_CONNECT_NBD_DURING_RESTORE'}) eq "false") {
		logmsg(LK::LK_INFO, '', $self->{'tag'}, 104228, 'NBD will be connected during quickCheck (see "LKDR_CONNECT_NBD_DURING_RESTORE" in /etc/default/LifeKeeper).');
	} else {
		$self->ResyncNBD();
	}
	$self->MirrorAvail(); # optional steps when mirror becomes available
	return 0;
FAIL:
	$self->MirrorUnavail(); # optional steps when mirror becomes unavailable
	return 1;
}

sub OutOfSync {
	my $source = shift;
	my $target = shift;

	# shared source is always in sync
	return 0 if (IsSharedSource($source, $target));
	# check if target (or target group) is in sync
	return 1 if ($target->{'percent'} < 100);

	return 0;
}

# App_remove
#   Description:  Called by the RKActionHandler::Remove method to do the
#     recovery kit specific remove work.
#   Arguments:  none.
#   Return Value:
#     0 if no errors, non-0 otherwise.
#
sub App_remove {
	my $self = shift;
	my $ret = 0;

	# in resize
	return 1 if ($self->is_in_resize('remove'));

	$self->PreRemove(); # optional pre-remove steps

	$self->UpdateStatus();

	if ($self->{'percent'} == 100) {
		$self->DeleteFlagFile('last_owner');
	}

	if ($self->{'percent'} == -1) { # mirror is not active
		goto OUT;
	}

	# if a remote resource is also ISP/ISU, avoid setting flags
	foreach ($self->SystemKeys()) {
		my $target = $self->Target($_);
		my $state = (split /\x01/, LK::ins_list($_, '', '', '', $target->{'tag'}, ''))[6];

		if ($state ne '' and $state ne 'ISP' and $state ne 'ISU') {
			# Allow user to remove the resource while NetRAID is not
			# mirroring, but set the appropriate flags and warn the
			# user that the NetRAID resource will have to be
			# brought ISP on another (in-sync) system first.
			#
			# Check the status by mirror number (for groups, we only
			# care if a mirror is established to any group member -
			# not necessarily the local system)
			# if there's a mirror target that is in sync anywhere in
			# the storage group, then all targets in the group are
			# considered to be in sync (since they're all shared)
			if (OutOfSync($self, $target)) {
				# Warn user
				logmsg(LK::LK_ERROR, 'remove', $self->{'tag'}, 104087, 'The mirror "%s" (resource: "%s") is not in sync between systems "%s" (source) and "%s" (target) (status: %s). You will not be able to failover the resource to the target system listed above. You will have to bring the resource in-service on another system first and a resync will occur to ensure data reliability', $self->{'md'}, $self->{'tag'}, $self->{'sys'}, $target->{'sys'}, $target->{'percent'});
				# Create appropriate flags
				if (!$target->CreateFlagFile('data_corrupt') && $target->{'state'} == $LK::ALIVE) {
					logmsg(LK::LK_ERROR, 'remove', $self->{'tag'}, 104218, 'Failed to create "data_corrupt" flag file on target "%s".', $target->{'sys'});
				}
				if (!$self->CreateFlagFile('last_owner')) {
					logmsg(LK::LK_ERROR, '', $self->{'tag'}, 104223, 'Failed to create "last_owner" flag file to track mirror source".  This may allow in-service of mirror on old data.');
				}
			} elsif (($self->{'percent'} == 100) && TargetSupportsWaitForSource($target->{'sys'})) {
				dbg("App_remove: All targets are in-sync, set \"source\" flag file on %s so that it does not require any previous source.", $target->{'sys'});
				if (!$target->CreateFlagFile('source', $target->{'sys'})) {
					logmsg(LK::LK_ERROR, 'remove', $self->{'tag'}, 104178, 'Failed to update "source" flag file on target "%s", previous source must be merged first.', $target->{'sys'});
				}
			}
		}
	}

	$ret = $self->StopNetRAID();
OUT:
	LK::flg_remove('', "!volatile!wait_to_resync_$self->{'tag'}");
	$self->MirrorUnavail(); # optional steps when mirror becomes unavailable
	return $ret;
}

# App_recover
#   Description: Called by the RKActionHandler::Recover() method to do the
#     recovery kit specific recover work.
#   Arguments: the failed device is passed in as -d argument (-d dev:$dev)
#              ("dev:" prefix avoids sendevent's zero-length -d argument error)
#   Return Value:
#     0 if no errors, non-0 otherwise.
#
sub App_recover {
	my $self = shift;
	my $dev = $1 if ($self->Arg('d') =~ /^dev:(.*)/);
	my $ret = 0;


	if (LockRecovery( $self->{'tag'} )) {
		logmsg(LK::LK_INFO, '', $self->{'tag'}, 104016, "Recovery for resource $self->{'tag'} already in progress.  Skipping.");
		return 0;
	}

	logmsg(LK::LK_INFO, '', '', 104017, "mirror $self->{'md'} (device: $dev)");

	# device can be NULL or "0:0" for some events, so check array status
	if ($dev eq '' || $dev eq '0:0') {
		$self->UpdateStatus();
		# status > -1 means NBD failed (but we don't know which one)
		if ($self->{'percent'} > -1) {
			$self->ResyncNBD(); # try to resync all, skip OK ones
			UnlockRecovery($self->{'tag'});
			return 0; # don't fail over
		}
	}
	if ($dev =~ m@^/dev/nb@) {
		$ret = $self->ResyncNBD($dev);
		$ret = 0; # don't fail over -- remote disk/system may be bad
	} else {
		$ret = $self->ResyncLocal();
	}

	UnlockRecovery($self->{'tag'});
	return $ret;
}

# App_delete
#   Description:  Called by the RKActionHandler::Delete method to do the
#     recovery kit specific delete work.
#   Arguments:  none.
#   Return Value:
#     0 if no errors, non-0 otherwise.
#
sub App_delete {
	my $self = shift;

	# cleanup the flag, mirrorinfo, log, and bitmap files for the resource
	$self->DeleteFlagFile('data_corrupt');
	$self->DeleteFlagFile('last_owner');	# in case it has been left around
	$self->DeleteFlagFile('startup_done');
	$self->DeleteFlagFile('source');
	$self->DeleteFlagFile('previous_source');
	LK::flg_remove('', "!volatile!wait_to_resync_$self->{'tag'}");
	DeleteMirrorInfoFile($self->{'num'});
	CleanLogs($self->{'bitmap'}, $self->{'target'}->{'rewindlog'},
		$self->{'target'}->{'templog'}, $self->{'target'}->{'tagfile'},
		MapFile($self->{'target'}->{'templog'}));

	# if unextend, then just unextend the netraid device
	if ($self->Arg('U')) {
		$self->TempUmount(); # undo temporary read/write target access
		# delete this system from all remote mirrorinfo files
		foreach ($self->SystemKeys()) {
			RemoveMirrorInfo($_, $self->{'num'}, $self->{'sys'});
		}
		# warn if resource is unextended during resync
		if ($self->FlagFileExists('data_corrupt')) {
			logmsg(LK::LK_ERROR, 'delete', $self->{'tag'}, 104088, 'Unextending resource "%s" while the mirror "%s" is out of sync. The data on the target partition may be corrupted', $self->{'tag'}, $self->{'md'});
		}

		$self->StopNetRAID(); # make sure dummy md, etc. is removed

		return $self->UnextendNetRAID();
	}

	# this is a normal delete -- nothing more to do if OSU or OSF
	return 0 if ($self->Val('InitState') ne 'AUTORES_ISP');

	# warn if resource is deleted while target(s) out of sync
	foreach my $sys ($self->SystemKeys()) {
		my $target = $self->Target($sys);
		if ($target->FlagFileExists('data_corrupt')) {
			logmsg(LK::LK_ERROR, 'delete', $self->{'tag'}, 104089, 'Deleting resource "%s" while target "%s" is out of sync. The data on the target partition may be corrupted', $self->{'tag'}, $sys);
		}
	}

	# update /etc/fstab to the local partition/mount point pair entry
	my $fstab_entry = `$ENV{'LKROOT'}/lkadm/bin/fstab -b "$self->{'md'}"`;
	if ($fstab_entry ne '') {
		my ($dev, $mp, $fs, $opt, $x, $fsck) = split ' ', $fstab_entry;
		`$ENV{'LKROOT'}/lkadm/bin/fstab -D -b "$self->{'md'}" -m "$mp"`;
		`$ENV{'LKROOT'}/lkadm/bin/fstab -A -b "$self->{'part'}" -m "$mp" -t "$fs" -k "$fsck" -o "$opt"`;
	}


	# lklin00004961 return 0 here.  In the case where the delete happens
	# without a remove, then StopNetRAID will fail, but the resource deletion
	# occurs anyway.  Returning 0 prevents lots and lots of un-needed
	# error messages on the gui.
	$self->StopNetRAID();
	return 0;
}

# Xen: verify VMs are running on the current mirror source -- if not, switchover
# (only do this for shared source systems)
sub PreCheck {
	my $self = shift;
	my $me = LK::lcduname();
	my $host;
	my $vms;

	return 1 if (! $self->IsXEN());

	chomp(my @info = `$NETRAID_BIN/mirror_status3 --terse $self->{'tag'}`);

	foreach (@info) {
		@_ = split /\x01/, $_;
		$vms = ($vms eq '') ? $_[0] : join ', ', ($vms, $_[0]);
		next if ($_[2] eq '');
		if ($host eq '') {
			$host = $_[2];
		} elsif ($host ne $_[2]) { # VMs running on multiple systems
			logmsg(LK::LK_INFO, '', $self->{'tag'}, 104018, "VMs ($vms) on mirror $self->{'tag'} running on multiple host systems. One or more VMs are UNPROTECTED. MANUAL INTERVENTION REQUIRED. Switch all VMs to the same host system to correct this problem.");
			return 0;
		}
	}
	if ($host ne '' and $host ne $me) { # switch mirror to host
		my $target = $self->Target($host);
		if (!defined($target)) {
			logmsg(LK::LK_INFO, '', '', 104019, "VMs ($vms) running on host $host but no mirror is configured. VMs are unprotected.");
			return 0;
		}
		# only for shared source systems (noone else has a dummy md)
		return 1 if (! $self->IsSharedSource($target));
		logmsg(LK::LK_INFO, '', $self->{'tag'}, 104020, "VMs ($vms) running on host $host but mirror source for $self->{'tag'} is $me. Switching mirror to $host.");
		system("lcdremexec -e -d \"$host\" -- perform_action -a restore -t $target->{'tag'}");
		return 0;
	}

	return 1; # no problems
}

# App_quickCheck
#   Description:  Called by the RKActionHandler::QuickCheck method to do the
#     recovery kit specific quickCheck work.
#   Arguments:
#	$ClassRef
#	$EventRef
#	$quickCheckYourAppRef
#   Return Value:
#     0 if no errors, non-0 otherwise.
#
sub App_quickCheck {
	my $self = shift;
	my ($ClassRef, $EventRef, $quickCheckYourAppRef) = @_;
	my $md_num = $self->{'num'};

	# in resize
	return 0 if ($self->is_in_resize('quickCheck'));

	# shared source setup
	foreach my $sys ($self->SystemKeys()) {
		my $target = $self->Target($sys);

		next if (!defined($target));

		if ($self->{'group'}) {
			if (IsSharedSource($self, $target)) {
				SetupSharedSource($target, $self->{'size'},
					$self->{'num'}, $self->{'numdisks'},
					$self->{'chunk'});
			} else {
				CleanupTarget($target);
			}
		}

	}

	# do precheck - return if it fails
	if (! $self->PreCheck()) { return 0; }

	# don't do any checking if there is no mirror
	if ($self->{'numdisks'} == 1) { return 0; }

	# Check if mdadm monitor is running
	if (! CheckProc('mdadm', $self->{'num'})) {
		StartProc('mdadm', $self->{'num'});
	}

	$self->UpdateStatus();
	if ($self->{'resync_pending'}) { # possible hung resync in PENDING state
		chomp(my $sync_action = `cat /sys/block/md$md_num/md/sync_action`);
		logmsg(LK::LK_WARN, 'quickCheck', $self->{'tag'}, 104156, 'Resynchronization of "%s" is in PENDING state. Current sync_action is: "%s"', $self->{'md'}, $sync_action);
		open(FD, ">/sys/block/md$md_num/md/sync_action");
		print FD "idle\n";
		close(FD);
		$self->{'resync_pending'} = 0;
	}

	# try to prevent hang in md_write_start (see bug 3681)
	open(FD, ">/sys/block/md$md_num/md/array_state");
	print FD "active\n";
	close(FD);

	if (IsUnsafeDevice($self->{'id'}, $self->{'part'})) { # check if DEVNAME is in use, and warn
		# PL-13245 - throttle down the message to one quickCheck per resource restore.
		my $throttleFlag = "!volatile!throttle_id_msg!$self->{'tag'}";
		if (LK::flg_test('', $throttleFlag) != 0) {
			LK::flg_create('', $throttleFlag);
			logmsg(LK::LK_EMERG, 'quickCheck', $self->{'tag'}, 104148, 'The storage configuration for mirror "%s (%s)" does not have a unique identifier and may have potential for data corruption in some environments in certain circumstances. Please refer to the SIOS Product Documentation for details on DataKeeper storage configuration options.', $self->{'tag'}, $self->{'id'});
		}
	}

	if ($self->{'percent'} < -1) { # local disk failure
		`sendevent -n "$self->{'id'}" -C netraid -E recover -d "dev:$self->{'part'}" -m LK:quickCheck 2>&1`;
		return 0;
	}
	foreach my $sys ($self->SystemKeys()) {
		my $target = $self->Target($sys);

		next if (!defined($target));

		# check for target (underlying) disk mounted (even when paused)
		if (TargetDeviceMounted($self, $target)) {
			$target->CreateFlagFile('data_corrupt');
			$target->DeleteFlagFile('last_owner'); # just in case
		}

		next if ($target->{'pause'});
		next if ($self->{'group'} and IsSharedSource($self, $target));

		my $num = $target->{'num'};
		my $pid = CheckProc('nbd-client', $num);

		# Verify that the LKDR network connection is active
		my $ipversion = LKIP::ip_get_version($target->{'ip'});
		my $ping_fail;

		if ($ipversion == '4') {
			$ping_fail = system("lkping -n -q -c $ENV{'LKDR_PING_NUMBER'} -w $ENV{'LKDR_PING_TIMEOUT'} $target->{'ip'} >/dev/null 2>&1");
		} elsif ($ipversion == '6') {
			#
			# lkping returns 0 on a "partial" success where less than 5 packets are sent and returned
			# before LKDR_PING_TIMEOUT expires.
			#
			# mainline ping and ping6 return 1 on a "partial" success, so we preload LKDR_PING_NUMBER
			# to make sure that we have an opportunity to get them all back before timing out.
			#
			$ping_fail = system("ping6 -n -q -l $ENV{'LKDR_PING_NUMBER'} -c $ENV{'LKDR_PING_NUMBER'} -w $ENV{'LKDR_PING_TIMEOUT'} $target->{'ip'} >/dev/null 2>&1");
		} else {
			logmsg(LK::LK_INFO, 'quickCheck', $self->{'tag'}, 104021, "Bad ipversion $ipversion for $target->{'ip'} !!!!!");
			$ping_fail = 1<<8; # shifted to match the return from system();
		}

		if ($ping_fail) {
			$target->{'pingfail'}++;
		} else {
			$target->{'pingfail'} = 0;
		}
		$self->SetPingFailCount($target->{'pingfail'}, $sys);

		if ($target->{'state'} == $LK::ALIVE) {
			next if (SplitBrain($self, $target));
		}

		# now that we know we're not in split brain or paused, do the target mirror active check
		if (TargetMirrorActive($self, $target)) {
			$target->CreateFlagFile('data_corrupt');
			$target->DeleteFlagFile('last_owner'); # just in case
		}

		if ($target->{'state'} != $LK::ALIVE or $ping_fail &&
		    $target->{'pingfail'} > $ENV{'LKDR_MAX_PING_FAIL'}) {
			if ($target->IsCurrentTarget()) {
				`mdadm --fail $self->{'md'} $target->{'nbd'} 2>&1`;
				`mdadm --remove $self->{'md'} $target->{'nbd'} 2>&1`;
				if ($pid) { # kill the client
					KillProc('nbd-client', $num);
					KillProc('balance', $num);
				}
			}

			my $date = now();
			my $reason = "ping failed $target->{'pingfail'} times (target: $sys)";
			if ($target->{'state'} != $LK::ALIVE) {
				$reason = "system $sys not alive";
			} elsif (!$target->CreateFlagFile('data_corrupt')) {
				logmsg(LK::LK_ERROR, 'quickCheck', $self->{'tag'}, 104208, 'Failed to create "data_corrupt" flag file on target "%s".', $target->{'sys'});
			}
			LK::ins_setstate($sys, $target->{'tag'}, $LK::OSU, "$reason ($date)");
			# log a message (only once)
			if ($target->{'pingfail'} == $ENV{'LKDR_MAX_PING_FAIL'} + 1) {
				logmsg(LK::LK_INFO, 'quickCheck', $self->{'tag'}, 104090, 'The replication connection for mirror "%s" (resource: "%s") is down (reason: %s)', $self->{'md'}, $self->{'tag'}, $reason);
			}
			next;
		}

		# check the status of the netraid device
		$self->UpdateStatus();
		if ($target->{'percent'} == 100 && $pid) { # make GUI see change
			`md_event SpareActive $self->{'md'} $target->{'nbd'}`;
		}

		next if ($target->{'percent'} >= 0 && $pid); # nbd is working

		# resync...
		`sendevent -n "$self->{'id'}" -C netraid -E recover -d "dev:$target->{'nbd'}" -m LK:quickCheck 2>&1`;
	}
	return 0;
}

# BitmapsSupported
#
# determine if the kernel supports bitmap files or not (currently 2.6.16+)
#
sub BitmapsSupported {
	my ($major, $minor, $rev) = (POSIX::uname())[2] =~ /^([0-9]+)\.([0-9]+)\.([0-9]+)/;
	return 0 if ($major < 2);
	return 0 if ($major == 2 && $minor < 6);
	return 0 if ($major == 2 && $minor == 6 && $rev < 16);
	return 1;
}

# Get resync status for /proc/mdstat
#
sub GetResyncStatus {
	my %syncdev;
	my $md;
	open(MDSTAT, "/proc/mdstat");
	@_ = <MDSTAT>;
	close(MDSTAT);

	foreach $_ (@_) {
		if ($_ =~ /^\s*$/) {
			$md='';
			next;
		}
		if ($_ =~ /^(md\d+)\s+:/) {
			$md=$1;
			next;
		}
		next if ($md eq '');
		if ($_ =~ /(recovery|resync|reshape|check)\s*= /) {
			# Resync in progess
			$syncdev{$md} = 1;
			next;
		}
	}
	return %syncdev;
}

# bug1347
# Exclusive resync mdadm wrapper (prevents simultaneous resyncs on 2+ arrays)
# This is a workaround for a kernel bug that is present in 2.6.18 (and others?)
# Returns <0: lock conflict, >0: mdadm failure, 0: success
sub AddRaidDev { # params: <--add|--re-add> /dev/md? [--write-mostly] /dev/nbd?
	my @argv = @_;
	my $ret = -1;
	open(LOCK, ">/tmp/sdr-add-dev") || return -1;
	# Exclusive lock prevents other processes from entering
	# this critical section
	flock(LOCK, Fcntl::LOCK_EX) || goto end;

	my %sync_devs = GetResyncStatus();
	$_ = keys %sync_devs;
	if ($_ == 0) {
		# no resync is going on -- add the device and resync this array
		chomp(my @out = `mdadm @argv --force 2>&1`);
		if ($? != 0) {
			$ret = $? >> 8;
			logmsg(LK::LK_ERROR, '', '', 104022, "$argv[1]: mdadm failed ($ret)");
			foreach (@out) {
				logmsg(LK::LK_ERROR, '', '', 104023, $_);
			}
		} else {
			$ret = 0;
		}
	}
end:
	close(LOCK);
	return $ret;
}

# ReAdd
#
# wrapper for mdadm --re-add/--add
# --add forces full resync (used during extend), --re-add uses the bitmap
#
# Note: enable write_mostly for sync arrays and async targets
# So sync targets in mixed sync/async arrays are the only ones w/o write_mostly,
# which means some reads may be issued over the network to those disks, but
# there currently is no way to specify write_mostly without also having async
# writes enabled for a given disk, so we have to settle for this
#
sub ReAdd { # params: md_dev device write_mostly re_add [system]
	my $md = shift;
	my $dev = shift;
	my $write_mostly = shift;
	my $re_add = shift;
	my $sys = shift;

	my $add = $re_add ? '--re-add' : '--add';
	my $wmostly = $write_mostly ? '--write-mostly' : '';
	my $me = LK::lcduname();

	if (!defined($sys) || $sys eq $me) {
		# local machine
		dbg("mdadm $add $md $wmostly $dev");
		my $ret = AddRaidDev($add, $md, $wmostly, $dev);
		return $ret if $ret > 0; # mdadm failure
	} else {
		# remote
		wrap($md, $dev, $wmostly, $re_add);
		`lcdremexec -d $sys -e -- /etc/default/LifeKeeper-perl -e '\" BEGIN { require \\"/etc/default/LifeKeeper.pl\\"; } use lib \\"$ENV{LKROOT}/lkadm/subsys/scsi/netraid/bin\\"; use RKActionHandler; use DRBase; exit DRBase::ReAdd($md, $dev, $wmostly, $re_add);\"'`;
		return $?;
	}
}

# check for split brain between self and target -- log and return 1 if so
sub SplitBrain { # params: self target
	my $self = shift;
	my $target = shift;

	return 0 if ($target->{'sys'} eq '');
	my $remins = LK::ins_list($target->{'sys'}, '', '', '', $target->{'tag'}, '');
	my $state = (split /\x01/, $remins)[6];
	if ($state =~ /^IS[PU]/) {
		my $msg;
		logmsg(LK::LK_EMERG, '', $self->{'tag'}, 104124, 'WARNING: A temporary communication failure has occurred between systems %s and %s. In order to avoid data corruption, data resynchronization will not occur. MANUAL INTERVENTION IS REQUIRED. In order to initiate data resynchronization, you should take one of the following resources out of service: %s on %s or %s on %s. The resource that is taken out of service will become the mirror target.', $self->{'sys'}, $target->{'sys'}, $self->{'tag'}, $self->{'sys'}, $target->{'tag'}, $target->{'sys'});
		return 1;
	}
	return 0;
}

# log a message about the target disk being mounted
sub LogTargetDeviceMounted { # params: local_tag target_sys target_partition
	my $tag = shift;
	my $target_sys = shift;
	my $target_part = shift;

	logmsg(LK::LK_EMERG, '', $tag, 104147, 'WARNING: The target system %s for mirror %s has the target disk %s currently mounted. In order to avoid data corruption, data resynchronization will not occur. MANUAL INTERVENTION IS REQUIRED. In order to initiate data resynchronization, you should unmount %s on %s. A full resync will occur.', $target_sys, $tag, $target_part, $target_part, $target_sys);
}

# do local and remote alerting about the target disk being mounted
sub AlertTargetDeviceMounted { # params: self target target_partition
	my $self = shift;
	my $target = shift;
	my $target_part = shift;

	LogTargetDeviceMounted($self->{'tag'}, $target->{'sys'}, $target_part);

	my ($tag, $sys, $part) = ($target->{'tag'}, $target->{'sys'}, $target_part);
	wrap($tag, $sys, $part); # use these only in the perl -e call below

	`lcdremexec -d $target->{'sys'} -e -- /etc/default/LifeKeeper-perl -e '\" BEGIN { require \\"/etc/default/LifeKeeper.pl\\"; } use lib \\"$ENV{LKROOT}/lkadm/subsys/scsi/netraid/bin\\"; use RKActionHandler; use DRBase; exit DRBase::LogTargetDeviceMounted($tag, $sys, $part);\"'`;
	`sendevent -n "$self->{'id'}" -C netraid -E tgtmount -d "dev:$target_part,sys:$target->{'sys'}" -m LK:quickCheck 2>&1`;
}

# check if the disk is mounted, return 1 if so, 0 otherwise
sub DeviceMounted { # params: sys disk
	my $sys = shift;
	my $disk = shift;

	if ($sys eq LK::lcduname()) {
		return LocalDeviceMounted($disk);
	}

	wrap($disk); # use this only in the perl -e call below

	# NOTE: This is kind of tricky, but we need lcdremexec failure to be distinct from a return code of 1
	# from LocalDeviceMounted, so flip the boolean twice. Failure of the lcdremexec and failure to find
	# the device mounted result in a non-zero exit code below, which we negate to get a 0 return code for
	# this function. Exit code of 0 means that the device was mounted, so negating that, we get a
	# return code of 1 for this function.
	`lcdremexec -d $sys -e -- /etc/default/LifeKeeper-perl -e '\" BEGIN { require \\"/etc/default/LifeKeeper.pl\\"; } use lib \\"$ENV{LKROOT}/lkadm/subsys/scsi/netraid/bin\\"; use RKActionHandler; use DRBase; exit (DRBase::LocalDeviceMounted($disk) == 0);\"'`;
	return (! $?);
}

# check if the disk is mounted locally, return 1 if so, 0 otherwise
sub LocalDeviceMounted { # params: disk
	my $disk = shift;

	return 0 if ($disk eq '');

	open(MTAB, "/etc/mtab");
	chomp(my @mtab = <MTAB>);
	close(MTAB);

	foreach my $entry (@mtab) {
		(my $mtab_disk = $entry) =~ s/^(\S+?)\s.*/\1/;

	CHECK:
		return 1 if ($disk eq $mtab_disk);

		# walk symlinks to be sure we're matching devices properly
		if (-l $mtab_disk) {
			$mtab_disk = Cwd::realpath($mtab_disk);
			goto CHECK; # check again
		}
	}
	return 0;
}

# check if the target (underlying) disk is mounted -- force a full resync, log, and return 1 if so
sub TargetDeviceMounted { # params: self target
	my $self = shift;
	my $target = shift;

	return 0 if ($target->{'sys'} eq '');

	my $id = (split /\x01/, LK::ins_list($target->{'sys'}, '', '', '', $target->{'tag'}, ''))[4];
	chomp(my $part = `getId -s $target->{'sys'} -p \"$id\"`);

	if (DeviceMounted($target->{'sys'}, $part)) {
		$self->SetFullSync($target->{'sys'}); # force a full resync
		AlertTargetDeviceMounted($self, $target, $part);
		return 1;
	}

	return 0;
}

sub LogTargetMirrorActive { # params: local_tag target_sys md
	my $tag = shift;
	my $target_sys = shift;
	my $md = shift;

	logmsg(LK::LK_EMERG, '', $tag, 104146, 'WARNING: The target system %s for mirror %s has the target mirror %s currently active. In order to avoid data corruption, data resynchronization will not occur. MANUAL INTERVENTION IS REQUIRED. In order to initiate data resynchronization, you should reboot system %s.', $target_sys, $tag, $md, $target_sys);
}

sub AlertTargetMirrorActive { # params: self target
	my $self = shift;
	my $target = shift;

	LogTargetMirrorActive($self->{'tag'}, $target->{'sys'}, $self->{'md'});

	my ($tag, $sys, $md) = ($target->{'tag'}, $target->{'sys'}, $self->{'md'});
	wrap($tag, $sys, $md); # use these only in the perl -e call below

	`lcdremexec -d $target->{'sys'} -e -- /etc/default/LifeKeeper-perl -e '\" BEGIN { require \\"/etc/default/LifeKeeper.pl\\"; } use lib \\"$ENV{LKROOT}/lkadm/subsys/scsi/netraid/bin\\"; use RKActionHandler; use DRBase; exit DRBase::LogTargetMirrorActive($tag, $sys, $md);\"'`;
	`sendevent -n "$self->{'id'}" -C netraid -E tgtmount -d "dev:$self->{'md'},sys:$target->{'sys'}" -m LK:quickCheck 2>&1`;
}

# check if the target mirror (md) device is active (and the mirror is not paused) -- log and return 1 if so
sub TargetMirrorActive { # params: self target
	my $self = shift;
	my $target = shift;

	return 0 if ($target->{'sys'} eq '');
	return 0 if ($target->{'pause'}); # if paused, it's normal for the md to be active

	my @active = GetMd($self->{'num'}, $target->{'sys'});

	if (@active) {
		AlertTargetMirrorActive($self, $target);
		return 1;
	}
	return 0;
}

# sort targets according to resource priority (highest priority first)
sub PrioritySort { # params: self @targets
	my $self = shift;
	my @targets = @_;
	my @result;

	my %prio = map { $self->{'eqv'}->{$_}->{'priority'} => $_ } @targets;

	@result = map { $prio{$_} } sort { $a <=> $b } keys %prio;

	return @result;
}

# ResyncLocal
#
# Description: handle local disk failure, based on default file setting:
#
#	LKDR_FAILURE=failover  # (default) perform a failover
#	LKDR_FAILURE=resync    # resync the local disk component
#	LKDR_FAILURE=nothing   # do nothing, MANUAL INTERVENTION REQUIRED
#
# Returns: 0 if no errors, non-0 otherwise
#
sub ResyncLocal {
	my $self = shift;

	# the local disk is now bad, do not allow restore
	if (!$self->CreateFlagFile('data_corrupt')) {
		logmsg(LK::LK_ERROR, '', $self->{'tag'}, 104209, 'Failed to create "data_corrupt" flag file on "%s".', $self->{'sys'});
	}
	$self->DeleteFlagFile('last_owner');

	# user has disabled failover and resync
	if (lc($ENV{'LKDR_FAILURE'}) eq 'nothing') {
		logmsg(LK::LK_EMERG, '', $self->{'tag'}, 104158, "WARNING: The local disk partition $self->{'part'} for data replication device\n$self->{'md'} has failed. MANUAL INTERVENTION IS REQUIRED.");
		return 0;
	}

	# by default, do a failover
	return 1 unless (lc($ENV{'LKDR_FAILURE'}) eq 'resync');

	# remove and re-add the local disk -- this will cause a resync
	system("mdadm --remove $self->{'md'} $self->{'part'} 2>/dev/null");
	return ReAdd($self->{'md'}, $self->{'part'}, 0, 1);
}




# ManageSDRS
# Restore/Remove the SDRS storage hierarchy resources on the specified server.
# Inputs are system name and our sdr resource tag name.
sub ManageSDRS {
	my $sys = shift;
	my $tag = shift;
	my $operation = shift;  # either remove or restore
	my $ret = 0;

	# get the sdrs resource tag name.
	my $sdrs = (split /\x01/, (LK::dep_list($sys, $tag, '', 'scsi', 'disrec'))[0])[1];
	if ($sdrs ne '') {
		# find all children of disrec and restore from the bottom
		my @deps = LK::dep_list($sys, $sdrs, '', '', '');
		foreach my $dep (@deps) {
			my $tag = (split("\x01", $dep))[1];
			$_ = `$ENV{'LKROOT'}/bin/lcdremexec -d "$sys" -- perform_action -b -t "$tag" -a $operation 2>/dev/null`;
			$ret = $?;
			dbg "perform_action $operation for $sys:$tag returns $ret";
			last if ($ret != 0);
		}
	}
	return $ret;

}

sub IsTypeXEN { # params: type
	return 1 if ($_[0] == 3);
	return 0;
}

sub IsTypeSDRS { # params: type
	return 0 if ($_[0] < 4 || $_[0] > 6);
	return 1;
}

sub IsXEN { # params: self
	my $self = shift;
	return IsTypeXEN($self->{'type'});
}

sub IsSDRS { # params: self
	my $self = shift;
	return IsTypeSDRS($self->{'type'});
}

# MoveTargetRes
# For SDRS type mirrors, move the disrec resources to the new target location.
# First, remove them from whereever they may be in-service, if not the new
# target location. Then, restore them on the new target location, if not
# there already.
sub MoveTargetRes {
	my ($self, $target, @possible_targets) = @_;

	return 0 if (!$self->IsSDRS());

	dbg "Bringing disrec resources in-service on $target->{'sys'} for group $target->{'group'}";
	# First remove SDRS where it's currently ISP.
	foreach my $sdrs_targ (@possible_targets) {
		my $sdrs_target = $self->Target($sdrs_targ);
		dbg "looking at $sdrs_target->{'sys'}, $sdrs_target->{'group'}";
		# skip this server if no sdrs tag
		next if (!defined($sdrs_target));
		# skip this server if it's us
		next if ($sdrs_target->{'sys'} eq $target->{'sys'});
		# skip this server if not part of the target's group
		next if ($sdrs_target->{'group'} ne $target->{'group'});

		# found a candidate - remove it there
		dbg "Removing disrec resources on $sdrs_target->{'sys'}";
		my $err = ManageSDRS($sdrs_target->{'sys'}, $target->{'tag'}, 'remove');
		dbg "Remove of disrec resources for $sdrs_target->{'sys'} returned $err";
	}

	# Now restore the SDRS hierarchy on the new target.
	dbg "Restoring target device resources on $target->{'sys'}";
	my $err = ManageSDRS($target->{'sys'}, $target->{'tag'}, 'restore');
	if ($err) {
		logmsg(LK::LK_ERROR, '', '', 104071, "Failed to restore target device resources on $target->{'sys'} : $err");
	}
	return $err;
}


# BitmapIsProtected
# For the given bitmap file determine if it is protected by a filesystem
# resource somewhere in the hierarchy beneath us.
sub BitmapIsProtected {
	my $self = shift;
	my $ret = 0;

	return 0 if (!$self->IsSDRS());

	# for each protected filesys in this hierarchy underneath our resource,
	# look for a match with the bitmap file path
	my @protFS = LK::ins_list('', $self->{'tag'}, 'gen', 'filesys', '', '');
	foreach (@protFS) {
		my $mntpt = (split("\x01", $_))[4];
		# look for a mountpoint match with our bitmap file.
		if ($self->{'bitmap'} =~ m@^$mntpt/@) {
			$ret = 1;
			last;
		}
	}
	return $ret;
}

# check for LK_INITDONE on sys - return true if found, false if not
sub LKInitDone { # params: sys (NULL is local)
	my $sys = shift;

	`lcdremexec -d \"$sys\" -- test -f $ENV{'LKROOT'}/config/LK_INITDONE 2>&1`;
	return (! $?);
}

# wait for resource initialization to complete - return false if we timeout
sub WaitForLKInitDone { # params: systems (NULL is local)
	my @sys_list = @_;

	foreach my $sys (@sys_list) {
		my $t = $ENV{'LKDR_RECOVER_TIMEOUT'} || 120;
		while (! LKInitDone($sys)) {
			dbg "wait for LK_INITDONE on $sys";
			sleep 5;
			return 0 if ($t-- == 0); # abort if flag doesn't appear
		}
	}
	return 1;
}

# this routine tries to recover mirrors after powerfail -- to be eligible,
# the mirror must be AUTORES_ISP and last owner locally, and no other system
# can be dead, last owner, AUTORES_ISP, or ISP
sub RecoverMirrors {
	my $dest = shift;
	my $me = LK::lcduname();

	my $pid = fork();
	return if ($pid > 0 || ! defined($pid));

	dbg "recover mirrors on sys $dest";
	if ($dest ne '' && $dest ne $me) {
		`lcdremexec -d $dest -e -- /etc/default/LifeKeeper-perl -e '\" BEGIN { require \\"/etc/default/LifeKeeper.pl\\"; } use lib \\"$ENV{LKROOT}/lkadm/subsys/scsi/netraid/bin\\"; use RKActionHandler; use DRBase; exit DRBase::RecoverMirrors($dest);\"'`;
		return;
	}

	# wait for resource initialization to complete, to avoid conflicts
	return if (! WaitForLKInitDone(''));

	my @ins = LK::ins_list('', '', 'scsi', 'netraid', '', '');
	my @autores = grep !/\001ISP\001/, grep /\001AUTORES_ISP\001/, @ins;
	AUTORES: foreach my $res (@autores) {
		@_ = split /\001/, $res;
		dbg "checking mirror $_[3] on $me";
		my $tag = $_[3];
		my @args = ('-t', $tag);
		my $self = new RKActionHandler('DRBase', $0, '', \@args);
		foreach my $sys ($self->SystemKeys()) {
			my $target = $self->Target($sys);
			next AUTORES if ($target->{'state'} != $LK::ALIVE);
			next AUTORES if ($target->FlagFileExists('last_owner'));
			my @targins = LK::ins_list($sys, '', '', '',
					$target->{'tag'}, '');
			my @targautores = grep /\001AUTORES_ISP\001/, @targins;
			next AUTORES if (@targautores);
			my @targisp = grep /\001ISP\001/, @targins;
			next AUTORES if (@targisp);
		}
		logmsg(LK::LK_INFO, '', '', 104024, "Restore AUTORES_ISP resource, last owner of mirror $tag");
		system("export SDR_FORCE_ONLINE=1; export SDR_FORCE_RELATED_CORRUPT_ONLINE=1; perform_action -a restore -t $tag");
	}
}

# SetFullSync
#   Description: This function will force a full resync to targetsys by
#	either dirtying the full bitmap or setting the fullsync flag in
#	the mirrorinfo file.
#   Arguments:
#	targetsys - Mirror target system (for setting the mirrorinfo file)
#   Return Value:
#     None
#
sub SetFullSync {
	my $self = shift;
	my $targetsys = shift;
	my $me = LK::lcduname();
	my $ret=0;
	my $bitmapbits=0;

	logmsg(LK::LK_INFO, '', $self->{'tag'}, 104219, "Set full resync to target %s.", $targetsys);
	if ($self->{'numdisks'} < 3) {
		# Bitmap merge is supported so determine the size of the bitmap
		# for dirtying the full bitmap.
		$self->UpdateStatus();
		$self->{'bitmap_dirty'} =~ /(\d+)\s+bits.*/;
		$bitmapbits = $1;
		# Dirty the full bitmap for single target configurations
		# unless the bitmap has no size in which fall through and
		# set the fullsync flag for the targetsys
		if ($bitmapbits > 0) {
			my $range = "0-" . ($bitmapbits - 1);
			$ret = bitmap_set_bits($self->{'num'}, $range);
			# if dirtying the full bitmap succeeded return,
			# otherwise fall through and set the fullsync
			# flag for targetsys.
			return if ($ret == 0);
			logmsg(LK::LK_WARN, '', $self->{'tag'}, 104145, 'Unable to dirty full bitmap.  Setting fullsync flag.');
		}
	}

	SetMirrorInfo($targetsys, $self->{'tag'}, $fullsync, 1, 1);
}

# ResyncNBD
#   Description: resync the specified NBD component or resync all NBDs
#   Arguments:
#      NBD device to resync (if NULL, attempt to resync all NBDs)
#      fullsync flag - if set, force a full resync
#      override_wait_to_resync flag - if set, connect NBD to sync
#             mirror even if resources are not in-service
#   Return Value:
#     0 if no errors, non-0 otherwise.
#
sub ResyncNBD {
	my $self = shift;
	my $nbd = shift;
	my $fullsync = shift;
	my $override_wait_to_resync = shift;
	my $ret = 0;
	my @targs;
	my $continue_waiting = 0;
	my $previous_source = '';

	$self->UpdateStatus();

	@targs = $self->SystemKeys();

	# for shared
	foreach my $targ (@targs) {
		my $target = $self->Target($targ);

		next if (!defined($target) or $target->{'sys'} eq '');

		if ($self->{'group'}) {
			if (IsSharedSource($self, $target)) {
				SetupSharedSource($target, $self->{'size'},
					$self->{'num'}, $self->{'numdisks'},
					$self->{'chunk'});
			} else {
				CleanupTarget($target);
			}
		}
	}

	# to avoid data corruption, do not start the resync if
	# a previous communication failure (split-brain) is detected
	# (i.e., both local and remote resources are ISP)
	foreach my $sys ($self->SystemKeys()) {
		my $target = $self->Target($sys);

		return 1 if (SplitBrain($self, $target));
	}

	if ($fullsync) {
		# stop waiting for the source if we are told to do a full sync
		$self->CleanupWaitForPreviousSource('');
	} else {
		($continue_waiting, $previous_source) = $self->ContinueToWaitForPreviousSource();
		return 0 if ($continue_waiting);
	}

	if (defined($nbd)) { # resync only the specified device
		$nbd =~ /^.*?(\d+)$/;
		my @list = $self->Targets($1);
		@targs = map { $_->{'sys'} } @list;
		if ($fullsync) {
			foreach my $targsys (@targs) {
				my $targobj = $self->Target($targsys);
				$self->SetFullSync($targobj->{'sys'});
			}
		}
	}

	# if there's already a resync in progress, just wait
	return 0 if ($self->{'percent'} > 0 && $self->{'percent'} < 100);

	# get the list of ALIVE and ready systems
	@targs = grep { $self->{'eqv'}->{$_}->{'state'} eq $LK::ALIVE } @targs;
	@targs = grep { LKInitDone($_) } @targs;

	return 0 if (!@targs);
	return 0 if (!$override_wait_to_resync && $self->WaitToConnectNBD());
	# need to remove the flag here so GUI will not show waiting to resync
	LK::flg_remove('', "!volatile!wait_to_resync_$self->{'tag'}");

	# remove the previous source
	@targs = grep { $_ ne $previous_source } @targs if ($previous_source ne '');
	@targs = $self->PrioritySort(@targs);

	foreach my $targ ($previous_source, @targs) {
		next if ($targ eq '');	# if no previous_source
		my $target = $self->Target($targ);

		next if (!defined($target) or $target->{'sys'} eq '');
		my $num = $target->{'num'};

		if ($target->{'pause'}) { # force GUI update to Paused
			LK::flg_create($target->{'sys'}, 'sdr_gui_tickle');
			LK::flg_remove($target->{'sys'}, 'sdr_gui_tickle');
			next;
		}

		# don't mirror to members of the same group
		if (IsSharedSource($self, $target)) {
			dbg "not mirroring to $target->{'sys'} in group $self->{'group'}";
			$target->DeleteFlagFile('previous_source');
			next if ($target->CreateFlagFile('source', $target->{'sys'}));
			logmsg(LK::LK_ERROR, '', $self->{'tag'}, 104170, "Failed to create \"source\" flag file on shared source %s to track mirror source. This may result in a full resync.", $target->{'sys'});
			next;
		}

		# to avoid data corruption, do not resync this target if:
		# 1) the underlying target device is mounted
		# or
		# 2) the target md device is active (this check must be after the split brain check (above),
		#    since a split brain mirror would meet this condition also)
		if (TargetDeviceMounted($self, $target) || TargetMirrorActive($self, $target)) {
			$target->CreateFlagFile('data_corrupt');
			$target->DeleteFlagFile('last_owner'); # just in case
			next;
		}

		my $dofull = $fullsync || $target->{'fullsync'};

		if ($dofull && $target->{'sys'} eq $previous_source) {
			# some how the mirror info for the previous source is set to full resync?
			# Mark all targets to full resync
			$self->CleanupWaitForPreviousSource($previous_source);
			$previous_source = '';
		}

		if (!$dofull) {
			# if not doing a full resync then check to see if this target requires a source we need to wait for
			my $targets_previous_source = $self->WaitForAnotherSource($target, $previous_source);
			if (!defined($targets_previous_source)) {
				$self->MarkTargetsToFullsync($targets_previous_source);
				$dofull = 1;
				$previous_source = '';
			} elsif ($targets_previous_source ne '') {
				$previous_source = $targets_previous_source;
				if (!$self->CreateFlagFile('previous_source', $targets_previous_source)) {
					logmsg(LK::LK_ERROR, '', $self->{'tag'}, 104174, "Failed to create \"previous_source\" flag file to track time waiting on source.  Will not be able to timeout.");
					logmsg(LK::LK_INFO, '', $self->{'tag'}, 104206, "To stop waiting for the previous source (forcing a full resync to remaining waiting targets) run: \"%s/bin/mirror_action %s fullresync %s %s\" on %s.",
						$ENV{'LKROOT'}, $self->{'tag'}, $self->{'sys'}, $target->{'sys'}, $self->{'sys'});
				}
			}
		}

		# make sure this is the previous source if we are still looking for one
		next if ($previous_source ne '' && $target->{'sys'} ne $previous_source);

		my $pid = CheckProc('nbd-client', $num);
		$self->UpdateStatus();
		my $percent = $target->{'percent'};

		dbg "resync(%d)... nbd$num: $percent $pid", $dofull;

		next if ($percent >= 1 && $pid); # don't try to resync twice

		if ($pid) { # another nbd-client is running... kill it
			logmsg(LK::LK_ERROR, '', $self->{'tag'}, 104091, 'Another nbd-client process is running for nbd%s (pid %s). This process will be terminated.', $num, $pid);
			KillProc('nbd-client', $num);
		}

		# SDRS - Move mirror target resources before creating new mirror
		if ($self->MoveTargetRes($target, @targs)) {
			$ret |= 1;
			logmsg(LK::LK_ERROR, '', $self->{'tag'}, 104092, 'Mirror target resource movement to system %s : status %s', $target->{'sys'}, 'Failed');
			next;
		}
		if ($self->IsSDRS()) {
			logmsg(LK::LK_INFO, '', $self->{'tag'}, 104093, 'Mirror target resource movement to system %s : status %s', $target->{'sys'}, 'Succeeded');
		}

		# Create the appropriate flag files
		if (!$target->CreateFlagFile('data_corrupt')) {
			logmsg(LK::LK_ERROR, '', $self->{'tag'}, 104210, 'Failed to create "data_corrupt" flag file on target "%s".', $target->{'sys'});
			next;
		}

		# Delete remote last_owner if it got left, e.g., after failover
		$target->DeleteFlagFile('last_owner');

		if (!$dofull) {
			my $ret = $self->MergeBitmap($target);
			if ($ret > 0) {
				$dofull = 1; # force full resync
				$self->CleanupWaitForPreviousSource($previous_source) if ($previous_source ne '');
				$previous_source = '';
			} elsif ($ret < 0) {
				# merge failed, will retry later
				return 0 if ($previous_source ne '');
				next;
			}
		}

		if (!$target->CreateFlagFile('source', $self->{'sys'})) {
			logmsg(LK::LK_ERROR, '', $self->{'tag'}, 104171, "Failed to create \"source\" flag file on %s to track mirror source. Target %s will not be added to mirror.", $target->{'sys'}, $target->{'sys'});
			next;
		}

		$target->DeleteFlagFile('previous_source');
		$self->DeleteFlagFile('previous_source') if ($previous_source ne '');
		$previous_source = '';

		$target->MirrorUnavail(); # mirror not usable on target

		my $server_args = $self->{'bitmap'} ? '' : '-a';
		# single thread with async mirror to avoid reordered writes
		$server_args .= $target->{'async'} ? ' -t1' : '';

		if ($target->{'rewindlog'} ne '') {
			$server_args .= " -l $target->{'rewindlog'}";
			if ($target->{'logmax'} =~ /[1-9]/) {
				$server_args .= " -L $target->{'logmax'}";
			}
		}
		my $id = (split /\x01/, LK::ins_list($target->{'sys'}, '', '', '', $target->{'tag'}, ''))[4];
		chomp(my $part = `getId -s $target->{'sys'} -p \"$id\"`);

		if (ConnectNBD($self->{'num'}, $self->{'sys'}, $num,
		    $target->{'sys'}, $target->{'ip'}, $part, $server_args,
		    $target->{'async'} || !$self->{'async'}, !$dofull, 0,
		    $target->{'compress'}) != 0) {
			logmsg(LK::LK_ERROR, '', $self->{'tag'}, 104094, 'Unable to start the resynchronization to system "%s" for mirror "%s" (resource: "%s"). Please check /proc/mdstat, the system logs, and LifeKeeper logs for error messages', $target->{'sys'}, $self->{'md'}, $self->{'tag'});
			$ret |= 1;
			next;
		}

		# set target resource OSU to trigger GUI update to Resync state
		my $date = now();
		LK::ins_setstate($target->{'sys'}, $target->{'tag'}, $LK::OSU, "resyncing ($date)...");

		my $full_or_partial = $dofull ? "Full" : "Partial";
		logmsg(LK::LK_INFO, '', $self->{'tag'}, 104095, '%s resynchronization of component "%s" has begun for mirror "%s"', $full_or_partial, "/dev/nbd$num", $self->{'md'});
	}
	return $ret;
}

# return: 1 if $sys supports "wait for source"
#         0 otherwise
sub TargetSupportsWaitForSource {
	my $sys = shift;

	my $full_version = `lcdremexec -e -d \"$sys\" -- rpm -q --qf '%{VERSION}-%{RELEASE}' steeleye-lkDR 2>/dev/null`;
	my ($version, $build) = split '-', $full_version;
	my ($major, $minor, $point) = split /\./, $version;

	if (($major > $wait_for_source_version{major}) or
	    ($major == $wait_for_source_version{major} and $minor > $wait_for_source_version{minor}) or
	    ($major == $wait_for_source_version{major} and $minor == $wait_for_source_version{minor} and $point > $wait_for_source_version{point}) or
	    ($major == $wait_for_source_version{major} and $minor == $wait_for_source_version{minor} and $point == $wait_for_source_version{point} and $build > $wait_for_source_version{build})) {
		return 1;
	}
	dbg("Target $sys does not support wait-for-source.  It is version %s, minimum wait-for-source version: %s.%s.%s-%s.", $full_version, $wait_for_source_version{major}, $wait_for_source_version{minor}, $wait_for_source_version{point},  $wait_for_source_version{build});
	return 0;
}

# return:  '' if not multi-target, assumes this was previous source
#          undef wait-for-source supported, missing file indicates corrupt
#          undef wait-for-source not supported and no corroboration that it was previous source
#          '' another server, probably new source, agrees this was the previous source
sub HandleMissingSourceFile {
	my $self = shift;
	my $target = shift;
	my $current_previous_source = shift;

	my $config = $self->GetClusterConfig();
	return '' if ($config < 2);

	if (TargetSupportsWaitForSource($target->{'sys'})) {
		logmsg(LK::LK_ERROR, '', $self->{'tag'}, 104212, "The \"source\" flag file on %s to track mirror source does not exist.  Full resync to remaining targets is required.", $target->{'sys'});
		return undef;
	}

	# The target is an older release where targets have a 'source' file but the source will not,
	# How do we know if this system was really the "last" source?  If the current_previous_source
	# points to this target then we have corroboration by either the new source or another target..
	return undef if ($current_previous_source ne $target->{'sys'});

	dbg("Target $target->{'sys'} does not have a source flag file, 'previous_source' flag file indicates it was previous source.  Merge bitmap with a partial resync.");
	return '';	# assume this was the previous source
}

# If the target needs to wait for another source
# then wait for that source
# If we have had a split brain where at some point there were
# multiple disjoint pairs lets try to wait for each source as
# needed by a target.
#
# return:
#   '' -  to not wait for another source
#   undef - do full resync
#  "previous source" - name of source to wait for
sub WaitForAnotherSource {
	my $self = shift;
	my $target = shift;
	my $current_previous_source = shift;

	my $previous_source = $target->ReadFlagFile('previous_source');
	$previous_source = $target->ReadFlagFile('source') if ($previous_source eq '');
	return($self->HandleMissingSourceFile($target, $current_previous_source)) if ($previous_source eq '');

	return '' if ($previous_source eq $self->{'sys'});	# current source was the previous source
	return '' if ($previous_source eq $target->{'sys'});	# Target was the previous source

	my $psrc = $self->Target($previous_source);
	if (!defined($psrc) or $psrc->{'sys'} eq '') {
		$target->DeleteFlagFile('source');	# new 'source' create later
		logmsg(LK::LK_ERROR, '', $self->{'tag'}, 104172, "The \"source\" flag file on %s does not contain a valid target (%s). Full resync to remaining targets is required.", $target->{'sys'}, $previous_source);
		return undef;	# force a full resync
	}

	return '' if (IsSharedSource($self, $psrc));	# target's previous source is shared with current source
	return '' if (IsSharedSource($target, $psrc));	# target's previous source is shared with this target
	my $num = $psrc->{'num'};
	return '' if (CheckProc('nbd-client', $num));	# already connected/merged

	# if previous_source ($psrc) is part of a shared storage group see if another system in the group is configured?
	foreach ($self->SystemKeys()) {
		my $sys = $self->Target($_);
		next if (!IsSharedSource($sys, $psrc));
		my $num = $sys->{'num'};
		return '' if (CheckProc('nbd-client', $num));	# configured
	}
	# previous source not ready.

	if ($ENV{'LKDR_WAIT_FOR_PREVIOUS_SOURCE_TIMEOUT'} == 0) {
		logmsg(LK::LK_INFO, '', $self->{'tag'}, 104015, "Do not wait for %s to merge bitmap, begin full resync to other targets.", $previous_source);
		return undef;
	}

	logmsg(LK::LK_INFO, '', $self->{'tag'}, 104184, "Target %s needs to wait for previous source %s to merge before it can do a partial resync.", $target->{'sys'}, $previous_source);
	return $previous_source;
}

# are source and target in same storage group (i.e., is target a shared source)?
sub IsSharedSource {
	my $source = shift;
	my $target = shift;

	return 0 if ($source->{'group'} eq '');
	return 0 if ($source->{'group'} ne $target->{'group'});
	return 1;
}

# Non-XenServer: just remove data_corrupt for any "Shared Source" nodes
# XenServer: set up a dummy array on "Shared Source" disks to allow XenMotion
# and other VM movement to another host in a Server Pool using XenCenter --
# this dummy array tracks writes that occur while we asynchronously switch the
# mirror to the new host that the VM is running on
sub SetupSharedSource {
	my $target = shift;
	my $size = shift;
	my $num = shift;
	my $num_disks = shift;
	my $chunk = shift;
	my $sys = $target->{'sys'};
	my $nbd_num = $target->{'num'};

	$target->DeleteFlagFile('data_corrupt');
	if (!$target->CreateFlagFile('source', $target->{'sys'})) {
		logmsg(LK::LK_ERROR, '', $target->{'tag'}, 104217, "Failed to create \"source\" flag file on shared source %s to track mirror source. This may result in a full resync.", $target->{'sys'});
		# FAIL?  I am not sure what good that would do at this point
	}

	my @ins = split /\x01/, LK::ins_list($sys, '', '', '', $target->{'tag'}, '');
	if ($ins[6] eq 'OSU') { # make sure GUI updates
		my $date = now();
		LK::ins_setstate($sys, $target->{'tag'}, $LK::OSU, "shared ($date)");
	}

	return if (! $target->IsXEN()); # Xen only

	my $bitmap = (split /\x02/, $ins[5])[8];
	my $id = $ins[4];
	chomp(my $disk = `getId -s $sys -p \"$id\"`);

	DummyMD($target, $size, $disk, $num, $num_disks, $bitmap, $chunk);
}

sub DummyMD {
	my $target = shift;
	my $size = shift;
	my $disk = shift;
	my $num = shift;
	my $num_disks = shift;
	my $bitmap = shift;
	my $chunk = shift;
	my $sys = $target->{'sys'};
	my $async = $target->{'async'};
	my $type = $target->{'type'};
	my $opts = undef;

	my $me = LK::lcduname();

	if ($sys eq $me) {
		return BuildArray($size, $disk, $num, $num_disks, $async, $bitmap, $chunk, $opts, 1, $type);
	}

	wrap($size, $disk, $num, $num_disks, $async, $bitmap, $chunk, $opts, $type);
	`lcdremexec -d $sys -e -- /etc/default/LifeKeeper-perl -e '\" BEGIN { require \\"/etc/default/LifeKeeper.pl\\"; } use lib \\"$ENV{LKROOT}/lkadm/subsys/scsi/netraid/bin\\"; use DRBase; exit DRBase::BuildArray($size, $disk, $num, $num_disks, $async, $bitmap, $chunk, $opts, 1, $type);\"'`;
	return ($? >> 8);
}

my $RAID_CHECK='/etc/sysconfig/raid-check';

sub UpdateRaidCheck {
	my $md = shift;
	my $add = shift;

	my @raid_check;

	return 0 if (! -f $RAID_CHECK);

	open my $NEWCONF, ">$RAID_CHECK.lk" || return 1;
	flock($NEWCONF, Fcntl::LOCK_EX) || goto FAIL;
	open CONF, "<$RAID_CHECK" || goto FAIL;
	@raid_check = readline CONF;
	close CONF;

	my @skip_devs = grep(/^SKIP_DEVS/, @raid_check);
	@raid_check = grep(!/^SKIP_DEVS/, @raid_check);
	my @devs;
	if (@skip_devs > 0) {
		$skip_devs[$#skip_devs] =~ /^SKIP_DEVS="(.*)"/;
		@devs = split(/ /, $1);
	}

	@devs = grep(!/$md$/, @devs);
	push @devs, $md if ($add);
	my $devs = join(' ', @devs);
	push @raid_check, "SKIP_DEVS=\"$devs\"\n";

	print $NEWCONF join('', @raid_check) || goto FAIL;
	IO::Handle::flush($NEWCONF) || goto FAIL;
	IO::Handle::sync($NEWCONF) || goto FAIL;
	rename "$RAID_CHECK.lk", $RAID_CHECK;
	close $NEWCONF;

	return 0;
FAIL:
	unlink "$RAID_CHECK.lk";
	close $NEWCONF;
	return 1;
}

sub BuildArray {
	my $size = shift;
	my $disk = shift;
	my $num = shift;
	my $num_disks = shift;
	my $async = shift;
	my $bitmap = shift;
	my $chunk = shift || $ENV{'LKDR_CHUNK_SIZE'} || $DEFAULT_BITMAP_CHUNK;
	my $opts = shift;
	my $target = shift; # is this a target-side (dummy) md?
	my $type = shift; # for target-side md only, the mirror type
	my $md = "/dev/md$num";
	my $first = $num * $MAX_TARGETS;
	my @devs;

	# need at least two disks (otherwise bitmap gets cleared)
	$num_disks = 2 if ($num_disks < 2);

	if ($target) {
		if (GetMd($num)) {
			dbg "$md already exists on target";
			goto OUT;
		}
		dbg "building $md on target";
		goto BUILD; # skip cleanup and build the array
	}

	# make sure there are no old daemon processes out there...
	KillProc('mdadm', $num);

	for (my $i = $first; !$target and $i < $first + $num_disks; $i++) {
		KillProc('nbd-server', $i); # kill nbd-servers running locally
		KillProc('nbd-client', $i);
		KillProc('balance', $i);
	}

	dbg "spawning monitor for $md";
	if (StartProc('mdadm', $num) != 0) {
		logmsg(LK::LK_ERROR, '', '', 104025, "failed to spawn monitor");
		goto FAIL;
	}

BUILD:
	# add nbds to prevent bitmap clearing - "missing" does not do this
	foreach ($first..$first + $num_disks - 2) {
		push @devs, "/dev/nbd$_";
	}

	if ($async == 1) {
		$opts .= " --write-behind";
	} elsif ($async > 1) {
		$opts .= " --write-behind=$async";
	}

	if ($bitmap) {
		$opts .= " --bitmap=$bitmap --bitmap-chunk=$chunk";
	}

	my @pid = DummyNBD($num, $size);

	my $build_cmd = "mdadm -B --auto=yes $md --level=1 --force $opts --raid-devices=$num_disks $disk @devs";
	dbg "creating $md: $build_cmd";
	chomp(my @out = `$build_cmd 2>&1`);
	if ($? != 0) {
		logmsg(LK::LK_ERROR, '', '', 104026, "cannot create $md:");
		foreach (@out) {
			logmsg(LK::LK_ERROR, '', '', 104027, $_);
		}
		KillProc('mdadm', $num);
		kill 'KILL', @pid;
		goto FAIL;
	}

	# remove dummy nbds (they were added to stop the bitmap getting cleared)
	foreach my $dev (@devs) {
		(my $nbdnum = $dev) =~ s/^.*?([0-9]+)$/$1/;
		system("mdadm $md --fail $dev 2>/dev/null");
		foreach (1..30) {
			system("mdadm $md --remove $dev 2>/dev/null");
			stat(devsysfs($num, "nbd$nbdnum")) or last;
			logmsg(LK::LK_INFO, '', '', 104072, "$md: Failed to remove $dev. Trying again... ($_)");
			sleep(2);
			if (!($_ % 5)) {
				logmsg(LK::LK_INFO, '', '', 104028, "$md: Still waiting for $dev to be removed");
			}
		}
	}
	kill 'KILL', @pid;

OUT:
	!$target and logmsg(LK::LK_INFO, '', '', 104029, "$disk is configured to be mirrored using $md");
	MirrorAvail($num, $type, '');
	return 0;
FAIL:
	$target and MirrorUnavail($num, $type, '');
	return 1;
}

# grow array to accomodate new target (not needed on first extend, numdisks == 2)
sub GrowArray { # params: sys md_num num_eqv
	my $sys = shift;
	my $md_num = shift;
	my $numdisks = shift;

	return if ($numdisks <= 2); # array always starts out with 2 slots

	# Note: 2.6.7 kernel is needed for this:
	`lcdremexec -e -d $sys -- "mdadm --grow --force /dev/md$md_num -n $numdisks"`;
	return ($? >> 8);
}

my $portmsg = "Port %s is in use on system %s. You might need\nto change the LKDR_BASE_PORT setting in /etc/default/LifeKeeper";

# start prog listening on sys:port with args [args = sprintf(argfmt, port)]
# retry up to tries times on error (retry on the next port if port is in use)
# return < 0 on error, return port number on success
sub TryListen { # params: tries sys port nbd_num prog argfmt
	my $tries = shift;
	my $sys = shift;
	my $port = shift;
	my $num = shift;
	my $prog = shift;
	my $argfmt = shift;
	my $ret;

	foreach my $try (1..$tries) {
		my $args = sprintf($argfmt, $port);
		$ret = StartProc($prog, $num, $args, $sys);
		last if ($ret == 0);
		if ($ret == 98) { # EADDRINUSE
			logmsg(LK::LK_INFO, '', '', 104030, $portmsg, $port, $sys);
			$port = NextPort($port);
		}
		logmsg(LK::LK_INFO, '', '', 104031, "Could not start $prog on $sys:$port (error $ret). Retrying...($try)");
	}
	return $ret ? -$ret : $port;
}

# set up compression tunnel from source to target:port
# returns: source port (> 0), or error (< 0)
sub Tunnel { # params: source target targ_ip nbd_num port compress
	my $source = shift;
	my $target = shift;
	my $ip = shift;
	my $num = shift;
	my $port = shift;
	my $compress = shift;

	return $port if ($compress == 0);

	my $tunnel = NextPort($port); # tunnel uses next higher port

	# tunnel from target:tunnel to target:port
	$tunnel = TryListen(2, $target, $tunnel, $num, 'balance',
				"-u$compress %s localhost:$port");
	return $tunnel if ($tunnel < 0);
	# tunnel from source:port to target:tunnel
	return TryListen(2, $source, $port, $num, 'balance',
				"-z$compress %s $ip:$tunnel");
}

sub KillTunnel { # params: source target nbd_num compress [silent]
	my $source = shift;
	my $target = shift;
	my $num = shift;
	my $compress = shift;
	my $silent = shift;

	return 0 if ($compress == 0);

	if (!$silent) {
		logmsg(LK::LK_INFO, '', '', 104032, "stopping network compression for /dev/nbd$num ...");
	}
	KillProc('balance', $num, $target);
	KillProc('balance', $num, $source);
}

sub IsComponentPresent { # params: md source dev
	my $md = shift;
	my $source = shift;
	my $dev = shift;
	`lcdremexec -d "$source" -- mdadm --detail $md | grep -qw $dev`;
	return !($?); # return 1 if grep returns 0, and vice versa
}

sub ConnectNBD { # params: md_num source nbd_num target target_ip target_part
		 #         server_args write_mostly re_add skip_md compress
	my $md_num = shift;
	my $source = shift;
	my $nbd_num = shift;
	my $target = shift;
	my $targ_ip = shift;
	my $targ_part = shift;
	my $server_args = shift;
	my $write_mostly = shift;
	my $re_add = shift; # do a re-add (partial resync)
	my $skip_md = shift; # skip md teardown and setup
	my $compress = shift; # network compression level

	my $ret = 0;
	my $md = "/dev/md$md_num";
	my $nbd = "/dev/nbd$nbd_num";

	if ($targ_part eq '') {
		logmsg(LK::LK_INFO, '', '', 104033, "Unable to determine the target device for the mirror $md.  The mirror will not be resumed.");
		return 6; # ENXIO
	}

	# Set the resync speed limit
	SetResyncSpeedLimit($source);

	# remove the nbd device from the array
	if (!$skip_md && IsComponentPresent($md, $source, $nbd)) {
		`lcdremexec -d "$source" -- mdadm $md --fail $nbd 2>&1`;
		`lcdremexec -d "$source" -- mdadm $md --remove $nbd 2>&1`;
		my $loop = 30;
		while ($? != 0 && IsComponentPresent($md, $source, $nbd)) {
			logmsg(LK::LK_INFO, '', '', 104034, "Failed to remove device $nbd from mirror $md...");
			sleep 2;
			`lcdremexec -d "$source" -- mdadm $md --remove $nbd 2>&1`;
			if ($loop-- == 0) {
				logmsg(LK::LK_ERROR, '', '', 104035, "Too many failures. Aborting resync of $md");
				return -16; # EBUSY
			}
		}
	}

	KillProc('nbd-server', $nbd_num, $target);
	KillTunnel($source, $target, $nbd_num, $compress, 1);

	my $port = FirstPort($nbd_num);

	# start nbd-server on the target system
	$port = TryListen(2, $target, $port, $nbd_num, 'nbd-server',
				"%s $targ_part $server_args");
	if ($port < 0) {
		logmsg(LK::LK_ERROR, '', '', 104036, "Failed to start nbd-server on $target (error $port)");
		return $port;
	}

	# set up compression tunnel, if necessary
	if ($compress) {
		$port = Tunnel($source, $target, $targ_ip, $nbd_num,
				$port, $compress);
		if ($port < 0) {
			logmsg(LK::LK_ERROR, '', '', 104037, "Failed to start compression (error $port)");
			return $port;
		}
		$targ_ip = 'localhost';
	}

	# start nbd-client on the source system
	foreach my $try (1..10) {
		$ret = StartProc('nbd-client', $nbd_num, "$targ_ip $port $nbd", $source);
		# retry to connection refused
		last if ($ret != 111);
		sleep 1;
	}
	if ($ret != 0) {
		logmsg(LK::LK_ERROR, '', '', 104038, "Failed to start nbd-client on $source (error $ret)");
		return $ret;
	}

	# now re-add the nbd device to the array
	if (!$skip_md) {
		$ret = ReAdd($md, $nbd, $write_mostly, $re_add, $source);
		if ($ret != 0) {
			logmsg(LK::LK_ERROR, '', '', 104039, "Failed to add $nbd to $md on $source");
			return $ret;
		}
	}
	return 0;
}

sub NBDInUse { # params: self nbd_num
	my $self = shift;
	my $nbd_num = shift;

	foreach ($self->{'sys'}, $self->SystemKeys()) {
		return 1 if ($self->Target($_)->{'num'} == $nbd_num);
	}
	return 0;
}

# StopNetRAID
#   Description:  Stop the NetRAID device.
#   Arguments:  none
#   Return Value:
#	0 if no errors, non-0 otherwise.
#
sub StopNetRAID {
	my $self = shift;
	my $loop = 0;

	logmsg(LK::LK_INFO, '', '', 104040, "stopping the monitor for $self->{'md'} ...");
	KillProc('mdadm', $self->{'num'});

	foreach (1..5) {
		open(PS, "ps h -fwC md_event |");
		my $process = grep /$self->{'md'}\D/, <PS>;
		close(PS);
		last if not $process;

		logmsg(LK::LK_INFO, '', '', 104041, "waiting for md_event to exit for $self->{'md'} ...($_)");
		sleep 2;
	}

	# if md device is mounted, then it will be left mounted
	open(MTAB, '/etc/mtab');
	my $mounted = grep /^$self->{'md'}\D/, <MTAB>;
	close(MTAB);
	if ($mounted) {
		logmsg(LK::LK_INFO, '', $self->{'tag'}, 104042, "$self->{'md'} will be left mounted");
		logmsg(LK::LK_INFO, '', $self->{'tag'}, 104096, 'Unable to stop/remove the mirror "%s" (resource: "%s") on system "%s". You can use the "mdadm --stop" command to manually stop the mirror', $self->{'md'}, $self->{'tag'}, $self->{'sys'});
		return 1;
	}

	open(MDSTAT, '/proc/mdstat');
	my $running = grep /^md$self->{'num'}\D/, <MDSTAT>;
	close(MDSTAT);
	while ($running) {
		logmsg(LK::LK_INFO, '', $self->{'tag'}, 104043, "stopping $self->{'md'} ...");
		chomp(my @out = `mdadm --stop $self->{'md'} 2>&1`);
		$running = $?;
		if ($running) {
			foreach (@out) {
				logmsg(LK::LK_INFO, '', $self->{'tag'}, 104044, $_);
			}
			if (++$loop >= $ENV{'LKDR_MAX_STOP_WAIT'}) {
				logmsg(LK::LK_ERROR, '', $self->{'tag'}, 104045, "failed to stop $self->{'md'}");
				logmsg(LK::LK_ERROR, '', $self->{'tag'}, 104097, 'Unable to stop/remove the mirror "%s" (resource: "%s") on system "%s". You can use the "mdadm --stop" command to manually stop the mirror', $self->{'md'}, $self->{'tag'}, $self->{'sys'});
				return 1;
			}
			sleep 5;
		}
	}

	foreach my $sys ($self->SystemKeys()) {
		my $num = $self->Target($sys)->{'num'};
		logmsg(LK::LK_INFO, '', $self->{'tag'}, 104046, "stopping the nbd-client for /dev/nbd$num ...");
		KillProc('nbd-client', $num);
		KillTunnel($self->{'sys'}, $sys, $num,
				$self->Target($sys)->{'compress'});
	}

	logmsg(LK::LK_INFO, '', $self->{'tag'}, 104047, "$self->{'md'} has been stopped");
	logmsg(LK::LK_INFO, '', $self->{'tag'}, 104098, 'The mirror "%s" (resource: "%s") has been successfully removed from system "%s"', $self->{'md'}, $self->{'tag'}, $self->{'sys'});

	return 0;
}


# break the nbd connection by killing the nbd-server locally (Target Side Only!)
sub UnextendNetRAID {
	my $self = shift;

	if (KillProc('nbd-server', $self->{'nbdnum'}) != 0) {
		logmsg(LK::LK_ERROR, '', $self->{'tag'}, 104099, 'Unable to unextend the mirror for resource "%s" from system "%s"', $self->{'tag'}, $self->{'sys'});
		return 1;
	}
	KillProc('balance', $self->{'nbdnum'});

	logmsg(LK::LK_INFO, '', $self->{'tag'}, 104100, 'The mirror for resource "%s" has been successfully unextended from system "%s"', $self->{'tag'}, $self->{'sys'});
	return 0;
}

# find an available md device number by checking /proc/mdstat
# and ins_list on all systems
sub FindAvailNetRAID_Num { # noparams
	my @syslist;
	my @inslist;
	my @mdout;
	my %mdhash;
	my $num = 0;

	@syslist = LK::sys_list('');
	foreach my $sys (@syslist) {
		@mdout = split /\n/, LK::lcdremexec($sys, "cat /proc/mdstat");
		foreach (@mdout) { /^md(\d+)\s/ and $mdhash{$1} = 1; }

		@inslist = LK::ins_list($sys, '', 'scsi', 'netraid', '', '');
		foreach (@inslist) {
			$_ = (split /\x01/, $_)[5]; # get the info field
			$_ = (split /\x02/, $_)[0]; # get the netraid number
			$mdhash{$_} = 1;
		}
	}

	while ($mdhash{$num}) {
		$num++;
	}
	return $num;
}


# Flag File Administration
# Use LOCAL TAG when creating, deleting, and checking on LOCAL SYSTEM,
# and REMOTE TAG when creating, deleting, and checking on REMOTE SYSTEM.

sub slashdot {
	my $val = $_[0];
	$val =~ s/\//./g;
	return $val;
}

# create flag file
# return 1 if file created, 0 otherwise
sub CreateFlagFile { # params: $self or $sys $tag
	my ($sys, $tag);

	# if called by an object, use its system and tag
	if (ref($_[0]) && $_[0] =~ /^\w+=HASH\(/) {
		my $self = shift;
		$sys = $self->{'sys'};
		$tag = $self->{'tag'};
	} else {
		$sys = shift;
		$tag = shift;
	}
	my $file = shift;
	my $memo = shift;
	$memo = '' if ($memo eq undef);
	$tag = slashdot($tag);

	logmsg2(LK::LK_INFO, '', $tag, 104159, "Create Flag $file (tag:$tag, sys:$sys)");
	my $tries = 2;
	my $ret = 1;
	while ($ret != 0 && $tries > 0) {
		$tries--;
		LK::lcderrclr();
		LK::lcdremexec($sys, "echo $memo > $NETRAID_ALARM/${tag}_$file");
		my $ret = LK::lcderrno();
		if ($ret == 0) {
			return 1 if ($memo eq '');
			# verify data in flag file
			LK::lcderrclr();
			my $data = LK::lcdremexec($sys, "cat $NETRAID_ALARM/${tag}_$file 2>/dev/null");
			chomp($data);
			return 1 if ($data eq $memo);
			dbg "Failed CreateFlagFile $file (tag:$tag, sys:$sys): data mismatch: expected $memo, found:$data.";
			LK::lcderrclr();
			LK::lcdremexec($sys, "rm -f $NETRAID_ALARM/${tag}_$file");
		} else {
			dbg "Failed CreateFlagFile $file (tag:$tag, sys:$sys): $ret.";
		}
	}
	return 0;
}

# delete flag file
sub DeleteFlagFile($$) {
	my $self = shift;
	my $file = shift;
	my $tag = slashdot($self->{'tag'});

	logmsg2(LK::LK_INFO, '', $tag, 104160, "Delete Flag $file (tag:$tag, sys:$self->{'sys'})");
	LK::lcdremexec($self->{'sys'}, "rm -f $NETRAID_ALARM/${tag}_$file");
}

# return last modification time for flag file (seconds since Epoch)
sub FlagFileModificationTime($$) {
	my $self = shift;
	my $file = shift;
	my $tag = slashdot($self->{'tag'});

	my $ret = `lcdremexec -d \"$self->{'sys'}\" -- stat -c %Y $NETRAID_ALARM/${tag}_$file 2>/dev/null`;
	return 0 if ($? != 0);
	chomp($ret);
	return $ret;
}

# check if flag file exists
sub FlagFileExists($$) {
	my $self = shift;
	my $file = shift;
	my $tag = slashdot($self->{'tag'});

	return system("lcdremexec -d \"$self->{'sys'}\" -- test -e $NETRAID_ALARM/${tag}_$file 2>/dev/null") == 0;
}

# read flag file contents
sub ReadFlagFile($$) {
	my $self = shift;
	my $file = shift;
	my $tag = slashdot($self->{'tag'});

	my $ret = `lcdremexec -d \"$self->{'sys'}\" -- cat $NETRAID_ALARM/${tag}_$file 2>/dev/null`;
	return '' if ($? != 0);
	chomp($ret);
	return $ret;
}

# return the next usable NBD number, given a starting point
sub NextNBD { # params: nbd_num
	my $nbdnum = shift;
	my $mirrornum = int($nbdnum) / $MAX_TARGETS;

	$nbdnum++;
	if ($nbdnum >= ($mirrornum + 1) * $MAX_TARGETS) { # wrap to first NBD
		$nbdnum = $mirrornum * $MAX_TARGETS;
	}
	return $nbdnum;
}

sub DummyNBD { # params: mirror_num dev_size
	my $num = shift;
	my $size = shift; # kilobytes unit
	$size *= 1024; # bytes unit

	my @c_pid;
	for (my $i = $num * $MAX_TARGETS; $i < ($num + 1) * $MAX_TARGETS; $i++) {
		my $dev = "/dev/nbd$i";
		my $port = FirstPort($num);
		my $s_pid;

		for (my $try = 0; $try < 100; $try++) {
			my $cmd="/usr/local/bin/nbd-server $port dummymode $size";
			chomp ($s_pid = `$cmd`);
			dbg "cmd: $cmd exitcode=$? pid=$s_pid";
			last if ($? == 0);
			$port = NextPort($port);
			$s_pid = -1;
		}
		return if ($s_pid == -1);
		my $cmd = "/usr/local/bin/nbd-client localhost $port $dev";
		chomp (my $pid = `$cmd`);
		dbg "cmd: $cmd exitcode=$? pid=$pid";
		if ($? != 0) {
			kill 'TERM',$s_pid;
		} else {
			push @c_pid, $pid;
		}
	}
	return @c_pid;
}

sub GetPartSize { # params: dev_name
	my $dev = shift;
	my $size;

	open(FD, "/proc/partitions") || return 0;

	$dev =~ s@^/dev/(.*)@$1@;

	while (<FD>) {
		@_ = split ' ';
		if ($_[3] eq $dev) {
			$size = $_[2];
			goto OUT;
		}
	}
OUT:
	close(FD);
	return $size;
}

# get the size of a device using the BLKGETSIZE ioctl
sub GetDevSize { # params: dev_name
	my $dev = shift;
	my $packed = chr(0) x 16;
	my $buf = \$packed;
        my $BLKGETSIZE64 = 0x80081272;

	my $arch = (POSIX::uname())[4];
	($arch =~ /^ppc/) and $BLKGETSIZE64 = 0x40041272; # PowerPC

	open(DEVICE, $dev) || return 0;
	my $ret = ioctl DEVICE, $BLKGETSIZE64, $buf;
	close(DEVICE);
	return GetPartSize($dev) if (!defined($ret)); # try /proc/partitions

	my @result = unpack "LL", $buf;
	my $dev_size = $result[0] + ($result[1] * 1.0 * 4294967296);
	$dev_size /= 1024; # ioctl returns bytes, we need kilobytes
	return int($dev_size);
}

# SetPingFailCount()
#   Description: Set the current ping fail count in the mirrorinfo file
#   Arguments:
#	the new ping fail count, to be stored in the local mirrorinfo file
#       the system to set the ping fail count for
#
sub SetPingFailCount($$$) { # params: pingfail system
	my $self = shift;
	my $value = shift;
	my $sys = shift;

	SetMirrorInfo($sys, $self->{'tag'}, $pingfail, $value, 1, 1);
}

sub SetPause($$$) { # params: pause system
	my $self = shift;
	my $value = shift;
	my $sys = shift;

	SetMirrorInfo($sys, $self->{'tag'}, $pause, $value, 1);
}

# SetResyncSpeedLimit
#    Description: Gets the resync speed limit value from the LK defaults file
#		 via the LKDR_SPEED_LIMIT environment variable, and then
#		 sets the resync speed limit value for currently active md
#		 devices via the proc filesystem interface.
#    Return Value: 0 if OK, 1 if errors
#
sub SetResyncSpeedLimit {
        dbg "SetResyncSpeedLimit: @_";
        if ($_[0] && $_[0] ne LK::lcduname()) {
                `lcdremexec -d $_[0] -e -- /opt/LifeKeeper/bin/perl -e '\"use lib \\"$ENV{LKROOT}/lib64/perl\\"; use lib \\"$ENV{LKROOT}/lkadm/subsys/scsi/netraid/bin\\"; use DRBase; exit DRBase::SetResyncSpeedLimit();\"'`;
        } else {
		$ENV{'LKDR_SPEED_LIMIT'} ||= 500000;
		$ENV{'LKDR_SPEED_LIMIT_MIN'} ||= 20000;
		if (system("echo $ENV{'LKDR_SPEED_LIMIT_MIN'} > $PROC_SPEED_LIMIT_MIN_PATH") != 0) {
			logmsg(LK::LK_ERROR, '', '', 104101, 'Failed to set the resynchronization speed limit in %s.', $PROC_SPEED_LIMIT_MIN_PATH);
			return 1;
		}
		if (system("echo $ENV{'LKDR_SPEED_LIMIT'} > $PROC_SPEED_LIMIT_MAX_PATH") != 0) {
			logmsg(LK::LK_ERROR, '', '', 104102, 'Failed to set the resynchronization speed limit in %s.', $PROC_SPEED_LIMIT_MAX_PATH);
			return 1;
		}
		return 0;
	}
}

our %NBD_PARAMS = ('scheduler' => 'deadline', 'rotational' => 1, 'nr_requests' => $ENV{'LKDR_ASYNC_LIMIT'} ? $ENV{'LKDR_ASYNC_LIMIT'} : 4096);

sub GetEnvOverride {
	my $param = shift;
	my $envvar = 'NBD_' . uc($param);

	return $ENV{$envvar}; # environment override
}

# set default block layer parameters for nbds to improve performance
sub SetBlockParams { # params: self
	my $self = shift;
	my $first = $self->{'num'} * $MAX_TARGETS;
	my $already = 0; # has the info message already been seen for this mirror?

	for (my $i = $first; $i < $first + $MAX_TARGETS; $i++) {
		while (my ($param, $value) = each (%NBD_PARAMS)) {
			my $file = "/sys/block/nbd$i/queue/$param";
			my $override = GetEnvOverride($param);
			$value = $override if (defined($override));
RETRY:
			my $err = 0;
			dbg "setting $file = $value\n";
			open(FD, '>' . $file) or $err = $!;
			print FD $value or $err = $!;
			close(FD) or $err = $!;

			if ($err) {
				# 4.x kernels have a limit of 2048 for nr_requests, so fall back and retry
				if ($param eq 'nr_requests' && $value > 2048) {
					if (!$already) {
						logmsg(LK::LK_INFO, '', '', 104226, 'Unable to set nr_requests to %s. Retrying with 2048.', $value);
						$already++;
					}
					$value = 2048;
					goto RETRY;
				}
				# 4.x kernels don't have "deadline" scheduler, so this is expected
				if ($param ne 'scheduler') {
					logmsg(LK::LK_ERROR, '', '', 104227, 'Failed to set %s to %s.', $file, $value);
				}
			}
		}
	}
}

# SetBitmap
#    Description: Sets the bitmap (if there is one) to clean or dirty
sub SetBitmap {
	my $self = shift;
	my $flag = shift;

	my $file = $self->{'bitmap'};
	return if (!$file);

	my $tag = $self->{'tag'};

	if ($flag eq 'clean') {
		system("bitmap -c \"$file\"");
		logmsg(LK::LK_INFO, '', $tag, 104103, 'Setting bitmap to clean for file %s (resource: "%s")', $file, $tag);
	} elsif ($flag eq 'dirty') {
		system("bitmap -d \"$file\"");
		logmsg(LK::LK_INFO, '', $tag, 104104, 'Setting bitmap to dirty for file %s (resource: "%s")', $file, $tag);
	} else {
		# error wrong flag setting
		logmsg(LK::LK_ERROR, '', $tag, 104105, 'Internal error in DR scripts has sent the wrong flag into SetBitmap. This is a non-fatal internal error.  Please report immediately to support@us.sios.com enclosing copies of the LifeKeeper log.');
	}
}

# return /proc/mdstat entry for /dev/md$num on $dest
sub GetMd { # params: num [dest]
	my $num = shift;
	my $dest = shift;
	my $md = "md$num";
	my $line;
	my @lines;
	my @mdstat;
	my $me = LK::lcduname();

	if ($dest && $dest ne $me) {
		chomp(@mdstat = `lcdremexec -d "$dest" -e -- cat /proc/mdstat`);
	} else {
		open (MDSTAT, '/proc/mdstat');
		@mdstat = <MDSTAT>;
		close MDSTAT;
	}

	for ($line = 0; exists $mdstat[$line]; $line++) {
		if ($mdstat[$line] =~ /^$md\s*:/) {
			chomp(@lines = @mdstat[$line .. $line+4]);
			last;
		}
	}

	return @lines;
}

sub SetTargetField { # params: self nbd_num field value
	my $self = shift;
	my $num = shift;
	my $field = shift;
	my $value = shift;

	foreach ($self->{'sys'}, $self->SystemKeys()) {
		my $thisnum = $self->Target($_)->{'num'};
		next if ($thisnum != $num);
		$self->{'eqv'}->{$_}->{$field} = $value;
	}
}

# see which remote IP the nbd-client is connected to (compression disabled) or
# see which remote IP the balance process is connected to (compression enabled)
sub FindTargetIP { # params: self nbd_num [system]
	my $self = shift;
	my $num = shift;
	my $dest = shift;

	foreach ('nbd-client', 'balance') {
		my $pid = CheckProc($_, $num, $dest);
		(my $ip, my $port) = GetProcIPInfo($pid, $dest);
		return $ip if ($ip ne '127.0.0.1' && $ip ne '::1'); # compression enabled
	}
}

# parse /proc/mdstat to update the following fields of $self (DRBase object)
#
#   'percent' - resync percentage [-2,-1,0,..,100]
#		-2 : Local disk failure
#		-1 : Mirror does not exist
#		 0 : Mirror is Broken
#	      1-99 : Mirror is Resynchronizing
#	       100 : Mirror is Mirroring
#
#   'speed' - resync speed in KB/s
#
#   'bitmap_dirty' - bitmap bits and dirty bits information:
#                    "Bitmap : 426 bits (chunks), 0 dirty (0.0%)"
#
#  also set up the target entries in 'eqv':
#
#   'percent' - resync percentage for this particular target
#		-1 : Target is not currently part of the mirror
#		 0 : Target is Out of Sync
#	      1-99 : Target is Resynchronizing
#	       100 : Target is Mirroring
#
# Arguments:
# dest - specifies the source system if it is other than the current system
#
sub UpdateStatus { # params: self [dest]
	my $self = shift;
	my $dest = shift;
	my $percent = -1;
	my $me = LK::lcduname();

	my @lines = GetMd($self->{'num'}, $dest);

	goto OUT if (@lines == 0);

	#
	# md0 : active raid1 nbd1[4](S)(W) nbd0[3](F) sda10[0]
	#
	# (F) = failed, (S) = spare, (W) = write-only (async)

	# start at the last device (e.g., sda10) and work backward
	while ($lines[0] =~ /^(.*)\s+([-\w]+?)(\d+)\[(\d+)\](\(*.*\)*).*?$/) {
		my $rest = $1;
		my $dev = $2;
		my $num = $3;
		my $slotmd = $4;
		if ($dev eq 'nbd') {
			# disk failed, treat it as if it's not in the array
			if ($5 =~ /F/) {
				$self->SetTargetField($num, 'percent', -1);
			} else {
				# newer kernels: mdstat doesn't have the right
				# slot number, so sysfs overrides if found
				my $f=devsysfs($self->{'num'}, "$dev$num", 'slot');
				my $slot = cat($f, $dest);
				$slot = $slotmd if (!defined($slot));
				$self->SetTargetField($num, 'slot', $slot);
				if ($self->{'group'}) {
					# which group member is current target?
					my $ip = $self->FindTargetIP($num, $dest);
					$self->SetTargetField($num, 'curtarget', $ip);
				}
			}
		}
		$lines[0] = $rest;
	}

	#
	#	208808 blocks super non-persistent [3/1] [U__]
	#
	$lines[1] =~ /^.*?\[(\d)\/(\d)\]\s*\[(.*)\]/;
	$self->{'disks'} = $1;
	$self->{'active'} = $2;
	$percent = ($1 == $2) ? 100 : 0;

	foreach ($me, $self->SystemKeys()) {
		next if (!defined($self->{'eqv'}->{$_}->{'slot'}));

		my $ok = $self->{'eqv'}->{$_}->{'slot'} < $self->{'disks'};
		$self->{'eqv'}->{$_}->{'percent'} = $ok ? 100 : 0;
	}

	my $local = (split //, $3)[0]; # check for "U"
	$percent = -2 if ($local ne 'U'); # local disk failure

	#
	#	[=====>...............]  resync = 28.9% (61008/208808) finish=0.3min speed=6778K/sec
	#	[=====>...............]  recovery = 28.9% (61008/208808) finish=0.3min speed=6778K/sec
	#         resync=DELAYED
	#
	if ($lines[2] =~ /resync\s*=/ || $lines[2] =~ /recovery\s*=/) {
		$self->{'resync_pending'} = (($lines[2] =~ /PENDING/)? 1 : 0);
		$percent = $1 if ($lines[2] =~ /re\w+\s+=\s+(\S+)\%/);
		$percent =~ s/\.\d+$//; # strip decimals
		$percent >= 1 or $percent = 1; # must be at least 1
		$self->{'speed'} = $1 if ($lines[2] =~ /speed=(.*)$/);
		my @detail = `lcdremexec -d "$dest" -e mdadm -D /dev/md$self->{'num'}`;
		foreach (@detail) {
			# Number Major Minor RaidDevice State       Device
			#    0    43      9      0      rebuilding  /dev/nbd9
			/^\s+\d+\s+(\d+)\s+\d+\s+\d+.*rebuilding.+\/dev\/nbd(\d+)/ or next;
			if ($1 == 43) { # nbd is major 43
				$self->SetTargetField($2, 'percent', $percent);
			}
			last;
		}
		# This status line appears as an extra line, so get rid
		# of it for reading bitmap status below
		shift @lines;
	}

	#
	#	bitmap: 0/2 pages [0KB], 64KB chunk, file: /tmp/bitmap
	#
	if ($lines[2] =~ /^\s*bitmap:.*, [0-9].* chunk, file: (.*)/) {
		if (-f $1) {
			$_ = `mdadm -X $1`;
			# Bitmap : 426 bits (chunks), 0 dirty (0.0%)
			/Bitmap.*:\s*(.*)/;
			$self->{'bitmap_dirty'} = $1;
		}
	}
OUT:
	$self->{'percent'} = $percent;
}

sub NextPort { # params: last_port
	return $_[0] + 256;
}

sub FirstPort { # params: mirror_num
	my $num = shift;
	my $port = ($ENV{'LKDR_BASE_PORT'} || 10001) + $num;

	open (FD, '/etc/services');
	my @services = <FD>;
	close(FD);

	while (grep { m@^[^#]+\s$port/tcp@ } @services) {
		$port = NextPort($port);
	}

	return $port;
}

# helper functions to start, stop, and check (using pid files in /var/run) the
# nbd-client, nbd-server, and mdadm monitor daemons that are used by LKDR

sub __kill_proc { # params: pid proc_name [signal]
	my $pid = shift;
	my $proc = shift;
	my $sig = shift || 15;

	dbg "__kill_proc $proc $sig $pid\n";
	if ($pid && system("ps $pid | grep -qw $proc") == 0) {
		kill $sig, $pid;
		return 1 if ($sig != 9 && $sig != 15); # non-fatal signal
		sleep 1;
		if (system("ps $pid >/dev/null 2>&1") == 0) {
			logmsg(LK::LK_WARN, '', '', 104048, "failed to kill $proc, pid $pid");
			return 1;
		}
	}
	return 0;
}

sub __pid_file { # params: proc_name mirror_num
	return "/var/run/$_[0].$_[1].pid";
}

# check the process and echo the pid, if the process is alive
# returns: PID if process is OK, 0 if process is not running
sub CheckProc { # params: proc_name mirror_num [system]
	dbg "CheckProc: @_";
	my $file = __pid_file($_[0], $_[1]);
	chomp(my $pid = `lcdremexec -d "$_[2]" -e -- "cat $file" 2>/dev/null`);
	return 0 if (!$pid);
	`lcdremexec -d "$_[2]" -e -- "ps $pid | grep -qw $_[0]" 2>/dev/null`;
	return 0 if ($? != 0);
	return $pid;
}

# start the process and save the pid, so it can later be checked or killed
sub StartProc { # params: proc_name mirror_num extra_args [system]
	dbg "StartProc: @_";
	my $cmd;
	my $user_args;
	my $errfile = '/tmp/sdr-start-proc.err.' . $$;

	if ($_[0] eq 'mdadm') {
		$cmd = "mdadm -F /dev/md$_[1] -p $ENV{'LKROOT'}/lkadm/subsys/scsi/netraid/bin/md_event -d 5 -f";
	} elsif ($_[0] eq 'nbd-client') {
		$cmd = '/usr/local/bin/nbd-client';
	} elsif ($_[0] eq 'nbd-server') {
		$cmd = '/usr/local/bin/nbd-server';
		$user_args = $ENV{'NBD_SERVER_ARGS'};
	} elsif ($_[0] eq 'balance') {
		$cmd = '/usr/local/bin/balance';
	}

	chomp (my $pid = `lcdremexec -d "$_[3]" -e -- "$cmd $_[2] $user_args" 2>$errfile`);
	my $ret = $? >> 8;
	dbg "start: $cmd $_[2] (sys:$_[3] ret:$ret pid:$pid)";
	if ($ret != 0) {
		logmsg(LK::LK_ERROR, '', '', 104125, "failed to start '$cmd $_[2] $user_args' on '$_[3]':");
		open(FD, "$errfile");
		foreach (<FD>) {
			chomp;
			logmsg(LK::LK_ERROR, '', '', 104126, $_);
		}
		close(FD);
	}
	unlink($errfile);
	return $ret if ($ret != 0);
	my $file = __pid_file($_[0], $_[1]);
	`lcdremexec -d "$_[3]" -e -- "echo $pid > $file"`;
	return 0;
}

# kill the process
sub KillProc { # params: proc_name mirror_num [system] [signal]
	dbg "KillProc: @_";
	if ($_[2] && $_[2] ne LK::lcduname()) {
		`lcdremexec -d $_[2] -e -- /etc/default/LifeKeeper-perl -e '\"use lib \\"$ENV{LKROOT}/lib64/perl\\"; use lib \\"$ENV{LKROOT}/lkadm/subsys/scsi/netraid/bin\\"; use DRBase; exit DRBase::KillProc(\\"$_[0]\\",$_[1],undef,$_[3]);\"'`;
	} else {
		my $pid = CheckProc($_[0], $_[1]);
		return 0 if ($pid == 0); #no pid or remexec failed, do not remove pid file, no pid found to kill
		my $sig;
		$sig = 9 if ($_[0] eq 'nbd-client'); # use SIGKILL for client
		$sig = $_[3] if ($_[3]);
		my $ret = __kill_proc($pid, $_[0], $sig);
		if ($ret == 0) { # remove pid file only if the process is dead
			unlink __pid_file($_[0], $_[1]);
		}
		return $ret;
	}
}

# returns array of remote IP, port to which the process is connected
sub GetProcIPInfo { # params: pid [server]
	my $pid = shift;
	my $server = shift;

	# Show only IPv4 and IPv6 connections
	chomp(my @out = `lcdremexec -d \"$server\" -e netstat -apn --inet --inet6 2>/dev/null | awk ' (\$7 ~ /^'$pid'\\//) { print \$5 }'`);
	foreach (@out) {
		# format is addr:port, e.g., 127.0.0.1:1234 (ipv4)
		# or 2002:4819:8549:1234:20da:5a8a:a456:1b0d:22 (ipv6)
		my ($ip, $port) = $_ =~ /^(.*):(.*?)$/;
		# balance has local and remote connections, skip the local
		next if ($ip eq '127.0.0.1' && @out > 1); # IPv4 localhost
		next if ($ip eq '::1' && @out > 1); # IPv6 localhost
		return ($ip, $port);
	}
}

my %_partinfo;
sub PartInfo { # params: partition
	my $part = shift;

	if (!defined(%_partinfo)) {
		open(FD, '/proc/partitions');
		while (<FD>) {
			@_ = split ' ', $_;
			$_partinfo{"/dev/$_[3]"} = HumanSize($_[2] * 1024) if ($_[2]);
		}
		close(FD);
	}
	if (!$_partinfo{$part}) {
		# for device mapper we need to use the /dev/dm-# name
		$_partinfo{$part} = $_partinfo{Cwd::realpath($part)};
	}
	return sprintf "$part%s\n", $_partinfo{$part} ? "  ($_partinfo{$part})" : "";
}

# manipulate the mirrorinfo files ($NETRAID_ALARM/mirrorinfo_$arraynum)
# there is one file per array that lists all targets (including the local sys)
# layout is (field separator is Control-A):
#
# 0          1          2         3     4        5     6      7      8
# target_sys target_num target_ip async pingfail pause logdir logmax fullsync
#
# 9        10    11
# compress group type
#
sub MirrorInfoFile { # params: <num or self>
	my $num;
	# if called by an object, use its md number
	if (ref($_[0]) && $_[0] =~ /^\w+=HASH\(/) {
		my $self = shift;
		$num = $self->{'num'};
	} else {
		$num = shift;
	}
	return "$NETRAID_ALARM/mirrorinfo_$num";
}


# parse the mirrorinfo file and create a hash indexed by system and nbd number
sub GetMirrorInfo { # params: <num or self> [dest]
	my $num;
	my @line;
	my %info;
	my $dest;

	# if called by an object, use its md number
	if (ref($_[0]) && $_[0] =~ /^\w+=HASH\(/) {
		my $self = shift;
		$num = $self->{'num'};
		$dest = shift;
		if (!defined($dest)) { # if no dest, then run on self
			$dest = $self->{'sys'};
		}
	} else {
		$num = shift;
		$dest = shift;
	}

	my $file = MirrorInfoFile($num);

	if ($dest && $dest ne LK::lcduname()) {
		@line = `lcdremexec -d "$dest" -e -- cat "$file"`;
	} else {
		open(FD, $file);
		@line = <FD>;
		close FD;
	}

	foreach (@line) {
		my %hash;

		chomp(@_ = split /\x01/, $_);
		$hash{'sys'} = $_[0];
		$hash{'num'} = $_[1];
		$hash{'md'} = "/dev/md$num";
		$hash{'nbd'} = "/dev/nbd$_[1]";
		$hash{'ip'} = $_[2];
		$hash{'async'} = $_[3];
		$hash{'pingfail'} = $_[4];
		$hash{'pause'} = $_[5];
		$hash{'logdir'} = $_[6];
		if ($_[6] ne '') {
			$hash{'rewindlog'} = LogName($_[6], $num);
			$hash{'tagfile'} = TagFile($hash{'rewindlog'});
			$hash{'templog'} = TempLogName($_[6], $num);
			$hash{'replaylog'} = ReplayLogName($_[6], $num);
		} else { # even without rewind, we use templog for r/w access
			$hash{'templog'} = TempLogName($ENV{'LKDR_TMPDIR'}, $num);
		}
		$hash{'logmax'} = $_[7];
		$hash{'fullsync'} = $_[8];
		$hash{'compress'} = $_[9];
		$hash{'group'} = $_[10];
		$hash{'type'} = $_[11];
		$hash{'percent'} = -1; # until we find out in UpdateStatus
		$info{$_[0]} = \%hash; # index by system
		dbg "get mirrorinfo $num: $_";
	}
	return \%info;
}

# wrap variables for remote execute (variables should not be used afterwards)
sub wrap {
	foreach (@_) {
		if (!defined($_)) { # pass as undef rather than empty string
			$_ = "undef";
		} else {
			$_ = "\\\"$_\\\"";
		}
	}
}

# update or delete a record (any fields that are NULL are preserved on update)
sub _SetMirrorInfo { # params: delete dest md
		     # sys num ip async pingfail pause logdir logmax fullsync
		     # compress group type
	my $delete = shift;
	my $dest = shift;
	my $md_num = shift;
	my ($sys, $num, $ip, $async, $pingfail, $pause, $logdir, $logmax, $fullsync, $compress, $group, $type) = @_;
	my $oldfile = MirrorInfoFile($md_num);
	my $newfile = "$oldfile.$$";
	my $lockfile = "$oldfile.lock";
	my $updated = 0;
	my $ret = 28; # ENOSPC
	my $me = LK::lcduname();

	if ($dest && $dest ne $me) {
		wrap($sys, $num, $ip, $async, $pingfail, $pause, $logdir, $logmax, $fullsync, $compress, $group, $type);
		`lcdremexec -d $dest -e -- /etc/default/LifeKeeper-perl -e '\" BEGIN { require \\"/etc/default/LifeKeeper.pl\\"; } use lib \\"$ENV{LKROOT}/lkadm/subsys/scsi/netraid/bin\\"; use DRBase; \\\$DRBase::logLevel_SetMirrorInfo = $logLevel_SetMirrorInfo; exit DRBase::_SetMirrorInfo(\\"$delete\\", \\"$dest\\", \\"$md_num\\", $sys, $num, $ip, $async, $pingfail, $pause, $logdir, $logmax, $fullsync, $compress, $group, $type);\"' 2>/dev/null`;
		return ($? >> 8);
	}

	logmsg2($logLevel_SetMirrorInfo, '', '', 104161, "set($delete,$me) mirrorinfo $md_num: $sys $num $ip $async $pingfail $pause $logdir $logmax $fullsync $compress $group $type");

	stack("invalid system") if ($sys eq '');

	open(LOCK, ">$lockfile") or return $!;
	flock(LOCK, Fcntl::LOCK_EX) or do { $ret = $!; goto OUT; };
	open(OLD, $oldfile);
	open(my $NEW, ">$newfile") or do { $ret = $!; goto OUT; };

	while (<OLD>) {
		logmsg2($logLevel_SetMirrorInfo, '', '', 104162, "set mirrorinfo $md_num: $_");
		if (! grep /^\Q$sys\E\x01/, $_) {
			print $NEW $_ or goto OUT;
		} elsif (!$delete) {
			chomp(my @old = split /\x01/, $_); # get existing values
			$num       = $old[1] if (!defined($num));
			$ip        = $old[2] if (!defined($ip));
			$async     = $old[3] if (!defined($async));
			$pingfail  = $old[4] if (!defined($pingfail));
			$pause     = $old[5] if (!defined($pause));
			$logdir    = $old[6] if (!defined($logdir));
			$logmax    = $old[7] if (!defined($logmax));
			$fullsync  = $old[8] if (!defined($fullsync));
			$compress  = $old[9] if (!defined($compress));
			$group     = $old[10] if (!defined($group));
			$type      = $old[11] if (!defined($type));
			print $NEW "$sys\x01$num\x01$ip\x01$async\x01$pingfail\x01$pause\x01$logdir\x01$logmax\x01$fullsync\x01$compress\x01$group\x01$type\n"
				or goto OUT;
			$updated = 1;
		}
	}
	if (!$updated && !$delete) {
		print $NEW "$sys\x01$num\x01$ip\x01$async\x01$pingfail\x01$pause\x01$logdir\x01$logmax\x01$fullsync\x01$compress\x01$group\x01$type\n"
			or goto OUT;
	}
	IO::Handle::flush($NEW);
	IO::Handle::sync($NEW);
	$ret = 0; # updates are done
OUT:
	close(OLD);
	close($NEW) or $ret = $!;
	if ($ret == 0) { # overwrite only if everything is OK
		$ret = ! rename($newfile, $oldfile); # returns 1 on success
	}
	flock(LOCK, Fcntl::LOCK_UN);
	close(LOCK);
	system("rm -f $newfile $lockfile");
	return $ret;
}

# add or update a record
sub AddMirrorInfo($$$$$$$$$$$$$$) { # params: dest md sys num
	# ip async pingfail pause logdir logmax fullsync compress group type
	return _SetMirrorInfo(0, @_);
}

# update the field_no'th field of a record (e.g., IP is field 2)
sub UpdateMirrorInfoField($$$$$) { # params: dest md sys field_no value
	my $field = $_[3];
	my $value = $_[4];
	my @a;
	$a[$field - 1] = $value;
	return _SetMirrorInfo(0, $_[0], $_[1], $_[2], @a);
}

# remove a record
sub RemoveMirrorInfo($$$) { # params: dest md_num sys
	return _SetMirrorInfo(1, @_);
}

# delete the whole file
sub DeleteMirrorInfoFile($) { # params: md_num
	my $file = MirrorInfoFile($_[0]);
	system("rm -f $file");
}

# copy mirrorinfo file for md_num from source to target
sub CopyMirrorInfo { # params: source md_num target
	my $source = shift;
	my $num = shift;
	my $target = shift;
	my $file = MirrorInfoFile($num);
	system("lcdremexec -e -d \"$source\" -- lcdrcp $file $target:$file");
}

# set mirrorinfo for sys on all eqv systems - returns: 0 success, 1 failure
sub SetMirrorInfo { # params: sys tag setting value [silent] [localonly]
	my $sys = shift;
	my $tag = shift; # the local tag
	my $setting = shift;
	my $value = shift;
	my $silent = shift;
	my $localonly = shift; # make the change locally only
	my $me = LK::lcduname();
	my @systems;

	if (!$localonly) {
		# settings need to be changed in mirrorinfo files on all servers
		@systems = map { split /\x01/; $_[2] } LK::eqv_list('', '', $tag);
	}
	push @systems, $me;

	my $info = (split /\x01/, LK::ins_list('', '', '', '', $tag, ''))[5];
	my $num = (split /\x02/, $info)[0];
	stack_die("invalid tag: $tag") if ($num eq '');

	my $field;
	my $name;

	if ($setting eq $pingfail) {
		$field = 4;
		$name = 'Ping Fail Count';
		$logLevel_SetMirrorInfo = LK::LK_DEBUG if ($value == 0);
	} elsif ($setting eq $pause) {
		$field = 5;
		$name = 'Pause Flag';
	} elsif ($setting eq $logdir) {
		$field = 6;
		$name = 'Rewind Log Directory';
	} elsif ($setting eq $logmax) {
		$field = 7;
		$name = 'Rewind Log Maximum Size';
	} elsif ($setting eq $fullsync) {
		$field = 8;
		$name = 'Full Resync Required';
		$logLevel_SetMirrorInfo = LK::LK_DEBUG if ($value == 0);
	} elsif ($setting eq $compress) {
		$field = 9;
		$name = 'Network Compression Level';
	} elsif ($setting eq $group) {
		$field = 10;
		$name = 'Storage Group';
	} elsif ($setting eq $type) {
		$field = 11;
		$name = 'Mirror Type';
	} else {
		stack_die("Invalid setting: $setting");
	}

	my $retval = 0;
	foreach my $dest (@systems) {
		if (!$silent) {
			logmsgout(LK::LK_INFO, '', $tag, 104049, "Setting $name for $sys to \"$value\" (on $dest)");
		}

		my $ret = UpdateMirrorInfoField($dest, $num, $sys, $field, $value);
		if ($ret != 0) {
			if (!$silent) {
				logmsgout(LK::LK_ERROR, '', $tag, 104050, "Setting $name on $dest failed: $ret. Please try again.");
			}
			$retval++;
		}
	}

	$logLevel_SetMirrorInfo = LK::LK_INFO;
	return $retval;
}

# the 'eqv' hash is keyed by system name
# delete the local system from the list so we iterate over targets only!
sub SystemKeys { # params: $self
	my $self = shift;
	my @list = grep !/^$self->{'sys'}$/, keys %{$self->{'eqv'}};
	return @list;
}

# get a given target from the eqv hash and bless it so it can act like a DRBase
# (since target hash has almost all the same fields as DRBase, most things work)
sub Target { # params: $self target_sys
	my $self = shift;
	my $target = shift;
	my $hash = $self->{'eqv'}->{$target};
	return undef if (!defined($hash));
	bless $hash, 'DRBase'; # let target object act like a DRBase object
	return $hash;
}

# get all targets from the eqv hash that match the given nbd number
# and bless those so they can act like DRBase objects
# (since target hash has almost all the same fields as DRBase, most things work)
sub Targets { # params: $self nbd_num
	my $self = shift;
	my $num = shift;
	my @list;

	foreach (keys %{$self->{'eqv'}}) {
		my $hash = $self->{'eqv'}->{$_};
		next if (!defined($hash) || $hash->{'num'} != $num);
		bless $hash, 'DRBase'; # let target object act like a DRBase
		push @list, $hash;
	}

	return @list;
}

sub IsCurrentTarget { # params: target
	my $target = shift;

	return 1 if ($target->{'group'} eq '');
	return 1 if (!defined $target->{'curtarget'});
	return 1 if ($target->{'curtarget'} eq $target->{'ip'});
	return 0;
}

sub IsMounted { # params: mount_point
	my $mp = shift;

	open(MTAB, '/etc/mtab');
	grep {
		@_ = split(' ');
		if ($mp eq $_[1]) {
			close(MTAB);
			return 1;
		}
	} <MTAB>;
	close MTAB;
	return 0;
}

# simple check to see if a device configuration is unsafe (i.e., DEVNAME is used in a non-LVM config)
# This check and the log message it generates is equivalent to collect_30_unsafedevname() in
# install/cd_layout/setuplibs/collect.sh
sub IsUnsafeDevice { # params: id partition
	my $id = shift;
	my $part = shift;

	return ($id eq $part && $id !~ m|^/dev/mapper|);
}

sub FindSource { # params: self
	my $self = shift;
	foreach ($self->SystemKeys()) {
		my $target = $self->Target($_);
		my $state = (split /\x01/, LK::ins_list($_, '', '', '', $target->{'tag'}, ''))[6];
		return $_ if ($state =~ /^IS[UP]$/);
	}
	return undef;
}

sub GetLocalPartition { # params: mountpoint
	my $MountPoint = shift;

	open(MTAB, '/etc/mtab') || logmsgdie('', '', 104051, 'Unable to open file "%s"', '/etc/mtab');
	my @mtab = grep /\s$MountPoint\s/, <MTAB>;
	close MTAB;
	@mtab == 1 || logmsgdie('', '', 104052, 'Specified existing mount point "%s" is not mounted', $MountPoint);

	@_ = split ' ', $mtab[0];
	my $LocalPartition = $_[0];
	my $fstype = $_[2];
	my @oldmountopts = split ',', $_[3];

	$LocalPartition = Cwd::realpath($LocalPartition);
	return ($LocalPartition, $fstype, @oldmountopts);
}

# rewind functions

my $logbasename = "rewindlog_";
sub LogName { # params: log_dir mirrornum
	return "$_[0]/$logbasename$_[1]";
}

my $templogbasename = "templog_";
sub TempLogName { # params: log_dir mirrornum
	return "$_[0]/$templogbasename$_[1]";
}

my $replaylogbasename = "replaylog_";
sub ReplayLogName { # params: log_dir mirrornum
	return "$_[0]/$replaylogbasename$_[1]";
}

sub TagFile { # params: log
	return ($_[0] . '.tags');
}

sub MapFile { # params: log
	return ($_[0] . '.map');
}

sub TempMount { # params: self [nomount]
	my $self = shift;
	my $nomount = shift;
	my $disk = $self->{'part'};
	my $log = $self->{'target'}->{'templog'};
	my $nbdnum = $self->{'nbdnum'};
	my $me = LK::lcduname();
	my $access = $self->{'md'};
	my $type = 'read/write';
	my $mountargs = '';

	if ($self->{'state'} =~ /^IS[UP]$/ || !$self->{'target'}->{'pause'}) {
		# temp mount is only allowed on target system with mirror paused
		logmsg(LK::LK_INFO, '', $self->{'tag'}, 104053, "Mirror $self->{'tag'} must be paused in order to enable access on the target");
		return 1;
	}

	my $source = $self->FindSource();

	# Temporarily mount the target disk for read-only or read/write access:
	# 1) if bitmap merging is available then R/W mount over local (dummy) md
	#      OR
	#    read-only mount over a local (loopback) nbd:
	#      1a) attach nbd-server with new data log (for throw-away data)
	#      1b) connect nbd-client to nbd-server on localhost
	# 2) if there's a filesys parent, mount to the filesys mount point

	# see if there's a filesys parent
	my $parent = Parent($self->{'tag'});
	my $fsid = ID($parent);

	# check for existing mount
	return 0 if (IsMounted($fsid));

	# with bitmap merging, we'll mount over md and merge the bitmaps to
	# throw away the target data when we're done

	my $ret = DummyMD($self->{'target'}, $self->{'size'}, $disk,
			$self->{'num'}, $self->{'numdisks'}, $self->{'bitmap'},
			$self->{'chunk'});
	goto OUT if ($ret != 0);

	return 0 if ($nomount); # nomount: skip filesys mount and messages below

	# if there's a filesys parent then mount it over the local md/nbd
	goto OUT if ($parent eq '');

	`$ENV{'LKROOT'}/subsys/gen/resources/filesys/actions/mountfs -d"$access" -t"$parent" -i"$fsid" $mountargs`;
	$ret = ($? >> 8);
	$access = $fsid; # access the filesystem through the mount point
OUT:
	if ($ret == 0) {
		logmsgout(LK::LK_INFO, '', $self->{'tag'}, 104054, "Temporary $type access for $self->{'tag'} set up on the mount point \"$access\" on server $me. You must access the data from this mount point ONLY! All changes made to the data on server \"$me\" via the mount point \"$access\" will be DISCARDED when the mirror is resumed.");
	} else {
		logmsgout(LK::LK_ERROR, '', $self->{'tag'}, 104055, "Failed to set up temporary $type access to data for $self->{'tag'}. Error: $ret");
		logmsgout(LK::LK_INFO, '', $self->{'tag'}, 104056, "Cleaning up, please wait...");
		`mdadm --stop $self->{'md'}`;
		$ret = ($? >> 8);
	}
	return $ret;
}

# undo the temporary mount done by TempMount
sub TempUmount { # params: self
	my $self = shift;
	my $nbdnum = $self->{'nbdnum'};
	my $is_rewind = shift;
	my $ret;

	return 0 if ($self->{'state'} =~ /^IS[UP]$/); # only allowed on target

	my $source = $self->FindSource();
	my $disk = "$self->{'md'}";

	# see if there's a filesys parent
	my $parent = Parent($self->{'tag'});
	my $fsid = ID($parent);

	# before unmounting, check that the target disk wasn't mounted by the user
	if (TargetDeviceMounted($self->Target($source), $self)) {
		$self->CreateFlagFile('data_corrupt');
		$self->DeleteFlagFile('last_owner'); # just in case
		return 1;
	}

	# unmount
	if ($parent ne '') {
		`$ENV{'LKROOT'}/subsys/gen/resources/filesys/actions/forceumount "$fsid" "$disk"`;
		$ret = ($? >> 8);
		if ($ret != 0) {
			logmsgout(LK::LK_ERROR, '', $self->{'tag'}, 104057, "Failed to undo temporary access for $self->{'tag'} on $self->{'sys'}. Error: $ret. Please verify that $fsid is not mounted on server $self->{'sys'}.");
			return $ret;
		}
	}
	if (GetMd($self->{'num'})) {
		`mdadm --stop $self->{'md'}`;
		$ret = ($? >> 8);
		if ($ret != 0) {
			logmsgout(LK::LK_ERROR, '', $self->{'tag'}, 104144, "Unable to stop the mirror access for $self->{'md'} on system $self->{'sys'}.  Error: $ret. Use the \"mdadm --stop $self->{'md'}\" command to manually stop the mirror.");
			return $ret;
		}
	}
	return 0;
}

sub ID { # params: tag
	return '' if ($_[0] eq '');
	return (split /\x01/, LK::ins_list('', '', '', '', $_[0], ''))[4];
}

sub Parent { # params: tag
	my $fs = (split /\x01/, (LK::dep_list('', '', $_[0], 'gen', 'filesys'))[0])[0];
	return $fs if ($fs);

	# otherwise go up the hierarchy until we find a filesys parent (if any)
	foreach (LK::dep_list('', '', $_[0], '', '')) {
		$fs = Parent($_);
		return $fs if ($fs);
	}
	return ''; # found none
}

# routine: ParentNotInService
# description:
#   Check if a parent resource is not in-service.  Parameters $parent and $state
#   are updated with the resource information for the resource that is not
#   in-service.
#   If no type is specified then check all parent resources, return the first
#   resource found that is OSF.  If no resource is OSF then return the first
#   resource found that is OSU.  Do not modify $parent/$state if all resources
#   are in-service or there are no parent resources.
#   If a type is defined then check hierarchy for all resources with that type,
#   return the first resource found that is OSF.  If no resource is OSF then
#   return the first resource found that is OSU.  Do not modify $parent/$state
#   if all resources of type are in-service or there are no resource of type.
#
# parameters:
#           tag - tag of the resource to start searching
#          type - the resource type, aka 'filesys' or '' to search all parents
#        parent - return the tag of the resource not in-service
#         state - return the state of the resource not in-service
#
# return: update $parent/$state with the tag/state of resource not ISP or ISU.
sub ParentNotInService {
	my ($tag, $type, $parent, $state) = @_;

	# check parents
	foreach my $dep (LK::dep_list('', '', $tag, '', '')) {
		my $ptag = (split /\x01/, $dep)[0];
		my ($restype, $pstate) = (split /\x01/, LK::ins_list('', '', '', '', $ptag, ''))[2,6];

		# check resource if $type is '' or this resource is the $type
		if ($type eq '' || $type eq $restype) {
			dbg "ParentNotInService tag=$tag, type=$type, parent tag=$ptag, state=$pstate.";
			if ($pstate eq 'OSF') {
				$$parent = $ptag;
				$$state = $pstate;
				return;
			} elsif ($$parent eq '' && $pstate eq 'OSU') {
				$$parent = $ptag;
				$$state = $pstate;
			}
		}
		# go up the hierarchy
		ParentNotInService($ptag, $type, $parent, $state);
		return if ($$state eq 'OSF');
	}
	return;
}

sub CleanLogs { # params: logs
	foreach (@_) {
		unlink($_);
	}
}

# strip trailing comment from a timestamp that was handed in from the GUI
# current separator is ": "
sub StripTimestampComment { # params: timestamp
	my $time = shift;

	if ($time =~ /^(.*?): /) {
		$time = $1;
	}
	return $time;
}

sub Disk { # params: tag
	my $tag = shift;

	my @ins = LK::ins_list('', '', '', '', $tag, '');
	my $id = (split /\x01/, $ins[0])[4];
	chomp(my $part = `getId -p \"$id\"`);
	return $part;
}

sub MirrorNum { # params: tag
	my $tag = shift;

	my @ins = LK::ins_list('', '', '', '', $tag, '');
	my $num = (split /\x02/, (split /\x01/, $ins[0])[5])[0];
	return $num;
}

sub GetMirrorInfoTag { # params: tag setting
	my $tag = shift;
	my $setting = shift;

	my $me = LK::lcduname();
	my $num = MirrorNum($tag);
	my $info = GetMirrorInfo($num);
	return $info->{$me}->{$setting};
}

sub LogFile { # params: tag
	my $tag = shift;
	return GetMirrorInfoTag($tag, 'rewindlog');
}

sub MyTime { # noparams
	(my $sec, my $usec) = gettimeofday();
	return $sec + ($usec / 1000000);
}

# returns the index and log file offset for an entry that matches timestamp
sub SearchTime { # params: log timestamp
	my $log = shift;
	my $time = shift;

	chomp(my $output = `logadm -l "$log" -s "$time"`);
	die "Unable to find timestamp \"$time\" in log\n" if ($output eq '');
	my ($index, $offset) = (split /\t/, $output)[0,4];
	return ($index, $offset);
}

# return logical log endpoints as (min_index, min_offset, max_index, max_offset)
sub GetEndpoints { # params: log [flags]
	my $log = shift;
	my $flags = shift;
	my ($min, $min_offset, $max, $max_offset);

	if ($flags) {
		$flags = "-f $flags";
	}

	chomp (my @output = `logadm -l $log $flags -e`);
	foreach (@output) { # there are two lines of output, min then max
		split /\t/;
		if (!defined($min)) {
			$min = $_[0];
			$min_offset = $_[4];
		} else {
			$max = $_[0];
			$max_offset = $_[4];
			last;
		}
	}

	dbg "endpoints: $min $min_offset $max $max_offset\n";
	return ($min, $min_offset, $max, $max_offset);
}

# end of rewind functions

1;
